""" Functions to import data from ANT Visor 2 / ANT EEG software go here """
import os
import h5py
import warnings
import numpy as np
import pandas as pd
from scipy import signal
from scipy.spatial.transform import Rotation

import pynibs

try:
    from pynibs.pckg import libeep
except (ImportError, SyntaxError):
    pass


def read_nlr(fname):
    """
    Reads NLR coordinates from *_recording.mri file.

    Parameters
    ----------
    fname : str
        FIle path of the NLR recording MRI file.

    Returns
    -------
    fiducials : np.ndarray of float
        (3, 3) The rows contain the fiducial points in ANT NIfTI space (nasion, left ear, right ear).
        Each fiducial point is represented as [x, y, z] coordinates.
    """
    f = open(fname, "r")
    text = f.readlines()

    fiducials = np.empty((3, 3))

    for i, line in enumerate(text):
        # nasion
        if "VoxelOnPositiveXAxis" in line:
            line = text[i + 1].replace("\t", " ")
            line = line.replace("\n", "")
            fiducials[0, :] = np.array([int(t) for t in line.split(" ")])

        # left ear
        if "VoxelOnNegativeYAxis" in line:
            line = text[i + 1].replace("\t", " ")
            line = line.replace("\n", "")
            fiducials[1, :] = np.array([int(t) for t in line.split(" ")])

        # right ear
        if "VoxelOnPositiveYAxis" in line:
            line = text[i + 1].replace("\t", " ")
            line = line.replace("\n", "")
            fiducials[2, :] = np.array([int(t) for t in line.split(" ")])

    return fiducials


def get_instrument_marker(im_path, verbose=False):
    """
    Return all instrument markers from visor .cnt file.

    Coordinate system in raw ANT space (NLR) is defined as:
    - origin: intersection between line of ear fiducials and nasion
    - x-axis: origin -> nasion
    - y-axis: origin -> left ear
    - z-axis: origin -> superior

    Parameters
    ----------
    im_path : str
        Path to instrument-marker-file .cnt file.
    verbose: bool, default: False
        Flag indicating verbosity.

    Returns
    -------
    im_list : list of dict
        List containing stimulation parameters.

        * coil_mean_raw: 4 x 4 numpy array
        * StimulusID: int
        * etc...

    Raises
    ------
    AssertionError
        If the .cnt file contains no instrument markers.
    """
    f = libeep.read_cnt(im_path)
    n_trig = f.get_trigger_count()
    # some triggers (2?) are some other information
    # so only take the ones with 'StimulusID' at 3rd position
    ims = [f.get_trigger(i)[3] for i in range(n_trig) if "StimulusID" in f.get_trigger(i)[3]]
    # or: if f.get_trigger(i)[0] == '6'

    if verbose:
        print(f"Found {len(ims)} instrument markers.")
    assert len(ims), "No instrument markers found in file"

    # now build list of matsimnibs from the instrument markers
    data = []

    for i, im in enumerate(ims):

        # transform string from .cnt file to dictionary
        data.append(dict(item.split('=') for item in im.split()[1:] if '=' in item))

        # floatify numeric variables
        for key in data[-1].keys():
            try:
                if key == "StimulusID":
                    data[-1][key] = int(data[-1][key])
                else:
                    data[-1][key] = float(data[-1][key])
            except ValueError:
                pass

        # transform to SimNIBS raw format
        matsimnibs_raw = np.zeros((4, 4))
        matsimnibs_raw[3, 3] = 1
        matsimnibs_raw[0:3, 3] = np.array([data[-1]['PosX'], data[-1]['PosY'], data[-1]['PosZ']]) * 1000
        quat = np.array([data[-1]['QuatX'], data[-1]['QuatY'], data[-1]['QuatZ'], data[-1]['QuatW']])
        matsimnibs_raw[0:3, 0:3] = Rotation.from_quat(quat).as_dcm()
        data[-1]["coil_mean_raw"] = matsimnibs_raw

    return data


def get_cnt_data(fn, channels='all', trigger_val='1', max_duration=10,
                 fn_hdf5=None, path_hdf5=None, verbose=False, return_data=False):
    """
    Reads ANT .cnt EMG/EEG data file and chunks timeseries into triggerN - trigggerN+1.

    It can directly write the zaps into hdf5 if argument is provided, starting with the first trigger and ending
    with get_sample_count()-1.

    Parameters
    ----------
    fn: str
        Path to the .cnt file.
    channels: str, int, list of int, or list of str, default: 'all'
        Which channel(s) to return. Can be channel number(s) or channel name(s).
    trigger_val: str, default: '1'
        Trigger value to read as zap trigger.
    max_duration : int, default: 10
        Maximum duration in [s] per chunk. Rest is dropped.
    fn_hdf5: str, optional
        If provided, the cnt data is written into an hdf5 file under "path_hdf5" as pandas dataframe
        with column name "qoi_name" and nothing is returned.
    path_hdf5: str, default: None
        If fn_hdf5, path within the HDF5 file where the data is saved (e.g. "/phys_data/raw/EEG")
    verbose: bool, default: False
        Flag indicating verbosity.
    return_data: bool, default: False
        If true, the data is returned as list of numpy arrays.

    Returns
    -------
    data_lst: list of np.ndarray, optional
        (samples,channels), List of EEG/EMG data. Only returned if "fn_hdf5" is not None.
    """

    f = libeep.read_cnt(fn)
    n_trig = f.get_trigger_count()
    n_samples = f.get_sample_count()
    n_channels = f.get_channel_count()
    sf = f.get_sample_frequency()
    chan_names = [f.get_channel(i)[0].lower() for i in range(n_channels)]

    if channels == 'all' or isinstance(channels, list) and channels[0] == 'all':
        channels_idx = range(n_channels)
    elif isinstance(channels, int):
        channels_idx = [channels]
    elif isinstance(channels, str):
        channels_idx = chan_names.index(channels.lower())
    elif type(channels) == list and all(type(chan) == int for chan in channels):
        channels_idx = channels
        assert np.all(np.array(channels_idx) >= 0), "Only positive channels numbers allowd"
        assert np.max(np.array(channels_idx)) < n_channels, f"Only {n_channels} channels found."

    elif type(channels) == list and all(type(chan) == str for chan in channels):
        channels_idx = [chan_names.index(chan.lower()) for chan in channels]

    else:
        raise NotImplementedError("Channels must be 'all', list(int), list(str)")

    assert channels_idx, "No channels with name / idx found."

    if fn_hdf5 is not None:
        assert path_hdf5, "Please provide path_hdf5="

    if verbose:
        print(f"get_cnt_data: {n_trig} triggers found.")
        print(f"get_cnt_data: {n_samples} samples found.")
        print(f"get_cnt_data: {sf} Hz sampling frequency.")
        print(f"get_cnt_data: {n_channels} channels found.")

    # get data between samples
    data_lst = []

    # chunk into triggers
    trigger_idx = 0
    # arr_idx = 0
    last_zap_done = False
    trigger_zap = 0
    # we want the data between trigger and trigger+1
    while trigger_idx < n_trig - 1:

        try:
            start = f.get_trigger(trigger_idx)

            # only use the triggers that have the correct trifger value
            if start[0] != trigger_val:
                if verbose:
                    print(f"get_cnt_data: Skipping idx {trigger_idx}: {start} (start)")
                trigger_idx += 1
                continue
            end = f.get_trigger(trigger_idx + 1)
            # also trigger+1 needs to have the correct trigger_val
            while end[0] != trigger_val:
                if verbose:
                    print(f"Skipping idx {trigger_idx}: {start} (end)")
                trigger_idx += 1
                if trigger_idx >= n_trig - 1:
                    break
                end = f.get_trigger(trigger_idx)

            # some sanity checks
            if not start[1] < end[1]:
                if verbose:
                    print(f"Trigger {trigger_idx} and {trigger_idx + 1}: wrong sample number "
                          f"({trigger_idx}: {start[1]}, {trigger_idx + 1}: {end[1]}]")
                # the eeg cnt files and with a trigger. get data from trigger to end-offile
                if trigger_idx == n_trig - 1:
                    end = (end[0], f.get_sample_count())
                    last_zap_done = True

            assert start[1] < (end[1] - 1), \
                f"Trigger {trigger_idx} and {trigger_idx + 1}: too close together " \
                f"({trigger_idx}: {start[1]}, {trigger_idx + 1}: {end[1]}]"

            # get sample number from trigger-tuple
            start = start[1]
            end = end[1] - 1
            length_org = end - start

            # cut to max duration chunk length
            end = np.min((end, start + sf * max_duration))

            if verbose:
                print(f"get_cnt_data: Trigger {trigger_idx:0>3}: {float(length_org) / sf:2.2}s / "
                      f"{float(end - start) / sf:0.2}s")
            data = f.get_samples(start, end)
            data_res = np.reshape(data, (end - start, n_channels), order='F')

            if return_data:
                data_lst.append(data_res)

        except (SystemError, UnicodeDecodeError) as e:
            print(f"Trigger {trigger_idx} error")
            print(e)
            continue

        if fn_hdf5 is not None:
            with h5py.File(fn_hdf5, "a") as fi:
                fi[path_hdf5 + f"/{trigger_zap:04d}"] = data_res
                trigger_zap += 1

        trigger_idx += 1

    # grap data for the last zap (trigger to end_of_file
    if not last_zap_done:

        try:
            start = f.get_trigger(trigger_idx)

            # only use the triggers that have the correct trigger value
            if start[0] != trigger_val:
                if verbose:
                    print(f"get_cnt_data: Skipping idx {trigger_idx}: {start} (start)")
                trigger_idx += 1
            end = f.get_sample_count()

            assert start[1] < (end - 1), \
                f"Trigger {trigger_idx} and {trigger_idx + 1}: too close together " \
                f"({trigger_idx}: {start[1]}, {trigger_idx + 1}: {end}]"

            # get sample number from trigger-tuple
            start = start[1]
            length_org = end - start

            # cut to max duration chunk length
            end = np.min((end, start + sf * max_duration))

            if verbose:
                print(f"get_cnt_data: Trigger {trigger_idx:0>3}: {float(length_org) / sf:2.2}s / "
                      f"{float(end - start) / sf:0.2}s")
            data = f.get_samples(start, end)
            data_res = np.reshape(data, (end - start, n_channels), order='F')

            if return_data:
                data_lst.append(data_res)

            if fn_hdf5 is not None:
                with h5py.File(fn_hdf5, "a") as fi:
                    fi[path_hdf5 + f"/{trigger_zap:04d}"] = data_res
                    trigger_zap += 1

            trigger_idx += 1

        except (SystemError, UnicodeDecodeError) as e:
            print(f"Trigger {trigger_idx} error")
            print(e)

        # reshape according to channel count
        # [chan1, chan2, chan3, chan1, chan2, chan3]
        # data_res = np.reshape(data, (end - start, n_channels), order='F')
        #
        # if return_data:
        #     data_lst.append(data_res[:, channels_idx])
        #
        # if fn_hdf5 is not None:
        #     with h5py.File(fn_hdf5, "a") as fi:
        #         fi[path_hdf5 + f"/{trigger_idx:04d}"] = data_res[:, channels_idx]

    # append last chunk
    # start = f.get_trigger(n_trig - 2)[1]
    # end = n_samples - 1
    # end = np.min((end, start + sf * max_duration))  # cut to max chunk length
    #
    # data = f.get_samples(start, end)
    # data_res = np.reshape(data, (end - start, n_channels), order='F')
    # if fn_hdf5:
    #     write_arr_to_hdf5(fn_hdf5=fn_hdf5,
    #                       arr_name=arr_name.format(arr_idx),
    #                       data=data_res[:, channels_idx],
    #                       verbose=verbose)
    # else:

    if return_data:
        return data_lst


def filter_emg(emg, fs):
    """
    Filter EMG signals.

    Parameters
    ----------
    emg : list of np.ndarray
        (n_stimuli), Raw EMG data. Each list entry contains a np.ndarray of size [n_samples x n_channel].
        Each channel is filtered in the same way.
    fs : float
        Sampling frequency.

    Returns
    -------
    emg_filt : list of np.ndarray
        (n_stimuli), Filtered EMG data.
    """

    # 5 Hz Butterworth high pass
    ############################
    b_butterhigh, a_butterhigh = signal.butter(N=5, Wn=5, btype='high', analog=False, fs=fs)
    # plot_frequency_response(a_butterhigh, b_butterhigh, fs=fs)

    # 200 Hz Butterworth low pass
    ############################
    b_butterlow, a_butterlow = signal.butter(N=5, Wn=200, btype='low', analog=False, fs=fs)
    # plot_frequency_response(a_butterlow, b_butterlow, fs=fs)

    # 50 Hz Notch filter
    ############################
    b_notch50, a_notch50 = signal.iirnotch(w0=50 / (fs / 2), Q=30)
    # plot_frequency_response(a_notch50, b_notch50, fs=fs)

    # 100 Hz Notch filter
    ############################
    b_notch100, a_notch100 = signal.iirnotch(w0=100 / (fs / 2), Q=50)
    # plot_frequency_response(a_notch100, b_notch100, fs=fs)

    # 150 Hz Notch filter
    ############################
    b_notch150, a_notch150 = signal.iirnotch(w0=150 / (fs / 2), Q=30)
    # plot_frequency_response(a_notch150, b_notch150, fs=fs)

    # 200 Hz Notch filter
    ############################
    b_notch200, a_notch200 = signal.iirnotch(w0=200 / (fs / 2), Q=30)
    # plot_frequency_response(a_notch200, b_notch200, fs=fs)

    # Filter signals
    emg_filt = []

    for e in emg:
        emg_filt.append(np.zeros(e.shape))
        for i_channel in range(e.shape[1]):
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_notch50, a_notch50, e[:, i_channel])
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_notch100, a_notch100, emg_filt[-1][:, i_channel])
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_notch150, a_notch150, emg_filt[-1][:, i_channel])
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_notch200, a_notch200, emg_filt[-1][:, i_channel])
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_butterlow, a_butterlow, emg_filt[-1][:, i_channel])
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_butterhigh, a_butterhigh, emg_filt[-1][:, i_channel])
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_notch50, a_notch50, emg_filt[-1][:, i_channel])

    return emg_filt


def merge_exp_data_visor(subject, exp_id=0, mesh_idx=0, verbose=False, start_mep=18, end_mep=35):
    """
    Merges all experimental data from visor experiment into one .hdf5 file.

    Parameters
    ----------
    subject : pynibs.Subject
        Subject object.
    exp_id : int, default: 0
        Experiment index.
    mesh_idx : int, default: 0
        Mesh index.
    verbose : bool, default: False
        Flag indicating verbosity.
    start_mep : float, default: 18
        Start of time frame after TMS pulse where p2p value is evaluated (in ms).
    end_mep : float, default: 35
        End of time frame after TMS pulse where p2p value is evaluated (in ms).

    Returns
    -------
    <File> : .hdf5 file
        File containing the stimulation and physiological data as pandas dataframes:

        * "stim_data": Stimulation parameters (e.g. coil positions, etc.)
        * "phys_data/info/EMG": Information about EMG data recordings (e.g. sampling frequency, etc.)
        * "phys_data/info/EEG": Information about EEG data recordings (e.g. sampling frequency, etc.)
        * "phys_data/raw/EMG": Raw EMG data
        * "phys_data/raw/EEG": Raw EEG data
        * "phys_data/postproc/EMG": Post-processed EMG data (e.g. filtered, p2p, etc.)
        * "phys_data/postproc/EEG": Post-processed EEG data (e.g. filtered, p2p, etc.)
    """
    # mep_paths_lst = subject.exp[exp_id]['fn_data']

    # im_lst = subject.exp[exp_id]['cond']
    # nii_exp_path_lst = subject.exp[exp_id]['fn_mri_nii']
    # nii_conform_path = subject.mesh[mesh_idx]['fn_mri_conform']
    fn_exp_hdf5 = subject.exp[exp_id]['fn_exp_hdf5']
    fn_current = subject.exp[exp_id]['fn_current'][0]
    # fn_coil = subject.exp[exp_id]['fn_coil']
    # fn_mesh_hdf5 = subject.mesh[mesh_idx]['fn_mesh_hdf5']
    exp_id = exp_id

    if os.path.exists(fn_exp_hdf5):
        os.remove(fn_exp_hdf5)

    # read stimulation parameters
    # ===================================================================================
    if 'fn_visor_cnt' in subject.exp[exp_id]:
        print(f"Reading stimulation parameters from {subject.exp[exp_id]['fn_visor_cnt']}")

        assert 'fn_fiducials' in subject.exp[exp_id]
        assert 'fn_current' in subject.exp[exp_id]
        assert len(subject.exp[exp_id]['fn_visor_cnt']) == 1, "Multiple coils not implemented for visor"
        fn_visor_cnt = subject.exp[exp_id]['fn_visor_cnt'][0]

        fn_fiducials = subject.exp[exp_id]['fn_fiducials'][0]

        ims_list = pynibs.visor.get_instrument_marker(fn_visor_cnt)
        ims_dict = pynibs.list2dict(ims_list)
        n_stim = len(ims_list)

        # read fiducials and transform to simnibs space
        fiducials = pynibs.visor.read_nlr(fn_fiducials)

        # fiducial correction
        if 'fiducial_corr' in subject.exp[exp_id]:
            fiducal_corr = np.array(subject.exp[exp_id]['fiducial_corr'])
            if any(np.abs(fiducal_corr[fiducal_corr != 0]) < .1):
                warnings.warn("fiducial_corr are expected to be given in mm.")
            fiducials += fiducal_corr

        fn_exp_nii = subject.exp[exp_id]['fn_mri_nii'][0][0]

        matsimnibs_raw = np.dstack(ims_dict["coil_mean_raw"])

        matsimnibs = pynibs.nnav2simnibs(fn_exp_nii=fn_exp_nii,
                                         fn_conform_nii=subject.mesh[mesh_idx]['fn_mri_conform'],
                                         m_nnav=matsimnibs_raw,
                                         nnav_system="visor",
                                         fiducials=fiducials,
                                         verbose=verbose)

        # read coil current
        current = np.loadtxt(fn_current)

        if subject.exp[exp_id]["cond"][0][0] != "":
            raise NotImplementedError("Individual conditions and average coil position over it not implemented yet")

        # create stim_data dataframe
        stim_data = {"coil_mean": [matsimnibs[:, :, i] for i in range(n_stim)],
                     "coil_type": [np.array(os.path.split(subject.exp[exp_id]["fn_coil"][0][0])[1]).astype(
                             "|S")] * n_stim,
                     "current": current,
                     "condition": [f"{(i - 1):04d}" for i in ims_dict["StimulusID"]]}

        df_stim_data = pd.DataFrame.from_dict(stim_data)
        df_stim_data.to_hdf(fn_exp_hdf5, "stim_data")

        print(f"Writing stim_data dataframe to {fn_exp_hdf5}")

    else:
        warnings.warn("No visor positions found.")

    # read emg
    # ===================================================================================
    if 'fn_emg_cnt' in subject.exp[exp_id]:

        print(f"Reading EMG data from {subject.exp[exp_id]['fn_emg_cnt'][0]}")

        # which emg_channel to use
        emg_channels = subject.exp[exp_id]['emg_channels']

        if isinstance(emg_channels, list) and len(emg_channels) > 1:
            warnings.warn("Multiple EMG channels are untested.")

        emg_trigger_value = subject.exp[exp_id]['emg_trigger_value'][0]

        max_duration = 10  # maximum EMG time series duration per after zap
        try:
            max_duration = subject.exp[exp_id]['emg_max_duration'][0]
        except KeyError:
            pass
        fn_emg_cnt = subject.exp[exp_id]['fn_emg_cnt'][0]

        # read info
        cnt_info = pynibs.get_cnt_infos(fn_emg_cnt)

        phys_data_info_emg = dict()
        for key in cnt_info.keys():
            phys_data_info_emg[key] = cnt_info[key]

        phys_data_info_emg["max_duration"] = max_duration
        phys_data_info_emg["emg_channels"] = emg_channels

        df_phys_data_info_emg = pd.DataFrame.from_dict(phys_data_info_emg)
        df_phys_data_info_emg.to_hdf(fn_exp_hdf5, "phys_data/info/EMG")
        print(f"Writing EMG info dataframe (phys_data/info/EMG) to {fn_exp_hdf5}")

        # read raw emg data from cnt file and write to hdf5 file
        emg = pynibs.visor.get_cnt_data(fn_emg_cnt,
                                        channels=emg_channels,
                                        max_duration=max_duration,
                                        trigger_val=emg_trigger_value,
                                        verbose=verbose,
                                        fn_hdf5=fn_exp_hdf5,
                                        path_hdf5="phys_data/raw/EMG",
                                        return_data=True)

        print(f"Writing EMG raw dataframe (phys_data/raw/EMG) to {fn_exp_hdf5}")

        # filter data
        emg_filt = pynibs.visor.filter_emg(emg=emg, fs=phys_data_info_emg["sampling_rate"])
        df_phys_data_postproc_emg = pd.DataFrame.from_dict({"filtered": emg_filt})

        # calc p2p
        # TODO: implement p2p function
        # p2p = calc_p2p(emg_filt)
        # df_phys_data_postproc_emg["p2p"] = p2p

        df_phys_data_postproc_emg.to_hdf(fn_exp_hdf5, "phys_data/postproc/EMG")
        print(f"Writing EMG postproc dataframe (phys_data/postproc/EMG) to {fn_exp_hdf5}")

    # read eeg
    # ===================================================================================
    if 'fn_eeg_cnt' in subject.exp[exp_id]:
        # which emg_channel to use?
        max_duration = 10  # maximum EMG time series duration per after zap

        try:
            max_duration = subject.exp[exp_id]['eeg_max_duration'][0]
        except KeyError:
            pass

        eeg_trigger_value = subject.exp[exp_id]['eeg_trigger_value'][0]

        # eeg_channel can be int, str, list of int, list of str
        eeg_channels = ['all']
        try:
            try:
                # list of int
                eeg_channels = subject.exp[exp_id]['eeg_channels']
            except ValueError:
                # list of str (gets casted to b'')
                eeg_channels = subject.exp[exp_id]['eeg_channels'].astype(str).tolist()
        except KeyError:  # key not defined, fall back to default
            pass

        fn_eeg_cnt = subject.exp[exp_id]['fn_eeg_cnt'][0]

        phys_data_info_eeg = dict()
        for key in cnt_info.keys():
            phys_data_info_eeg[key] = cnt_info[key]

        phys_data_info_eeg["max_duration"] = max_duration
        phys_data_info_eeg["eeg_channels"] = eeg_channels

        df_phys_data_info_eeg = pd.DataFrame.from_dict(phys_data_info_eeg)
        df_phys_data_info_eeg.to_hdf(fn_exp_hdf5, "phys_data/info/EEG")
        print(f"Writing EEG info dataframe (phys_data/info/EEG) to {fn_exp_hdf5}")

        # read raw eeg data from cnt file and write to hdf5 file
        pynibs.visor.get_cnt_data(fn_eeg_cnt,
                                  channels=eeg_channels,
                                  max_duration=max_duration,
                                  trigger_val=eeg_trigger_value,
                                  verbose=verbose,
                                  fn_hdf5=fn_exp_hdf5,
                                  path_hdf5="phys_data/raw/EEG",
                                  return_data=False)

        print(f"Writing EEG raw dataframe (phys_data/raw/EEG) to {fn_exp_hdf5}")

    print("DONE")



import scipy
import numpy as np
from sklearn.neighbors import KernelDensity

import pynibs


def mc(idx_list, array, ele_idx_1, mode="cols", **kwargs):
    """
    Determines mutual coherence for given zap indices in idx_list.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : np.ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements.
    ele_idx_1 : np.ndarray of float [n_ele]
        Element indices for which the optimization is performed.
    mode : str, default: "cols"
        Set if the mutual coherence is calculated w.r.t. columns or rows ("cols", "rows").

    Returns
    -------
    res : np.ndarray of float [n_combs]
        Mutual coherence. Lower values indicate more orthogonal e-field combinations (better).
    """

    res = []
    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    for ind in idx_list:
        if mode == "cols":
            res.append(pynibs.mutual_coherence(array[ind, :]))
        elif mode == "rows":
            res.append(pynibs.mutual_coherence(array[ind, :].transpose()))
        else:
            raise NotImplementedError("Specified mode not implemented. Choose 'rows' or 'cols'.")

    return res


def roi_elmt_wise_corr(idx_list, e_mat, ele_idx_1, decorrelate_hotspot_only=False, backend=np, **kwargs):
    """
    Compute element wise correlation for sets of e-field.

    Parameters
    ----------
    idx_list : list of int
        List of efield indicies in e_mat.
    e_mat : np.ndarray
        All e-fields.
    ele_idx_1 : list of int
        All element indices to compute corcoeff for.
    decorrelate_hotspot_only : bool, default: False
        If true, ele_idx_1[-.1] is used for decorrelation.
    backend : module
        Package to use to compute correlation - probably either numpy or cuda.

    Returns
    -------
    res : np.ndarray
        (e_mat.shape[0], ) with correlations for all e-field sets in idx_list.
    """
    res = backend.zeros(len(idx_list))

    for result_list_idx, coil_conf_seq_idcs in enumerate(idx_list):
        # extract values of interest from e-field matrix
        e_selected_subsampled = e_mat[coil_conf_seq_idcs][:, ele_idx_1]

        # compute column-wise correlation: we want to reduce the e-field correlation between ROI elements
        correlation_mat = backend.corrcoef(e_selected_subsampled, rowvar=False)

        if decorrelate_hotspot_only:
            # assumption: hotspot index is always included at last position of 'ele_idx_1'
            res[result_list_idx] = backend.mean(backend.abs(correlation_mat[-1, :]))
        else:
            # temp_mat = backend.abs(correlation_mat)
            # res[result_list_idx] = backend.mean(correlation_mat[backend.triu_indices_from(temp_mat, k=1)])
            res[result_list_idx] = backend.mean(backend.abs(correlation_mat))

        # num_nonzero_elmts = backend.count_nonzero(temp_mat)
        # matrix score
        # score = 1 / num_nonzero_elmts * backend.sum(temp_mat)
    return res


def coil_wise_corr(idx_list, array, ele_idx_1, **kwargs):
    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    res = []

    for ind in idx_list:
        r = np.corrcoef(array[ind, :])
        res.append(np.mean(r[np.triu_indices(r.shape[0], k=1)]))

    return res


def var(idx_list, array, ele_idx_1, **kwargs):
    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    res = []

    for ind in idx_list:
        res.append(np.mean(np.var(array[ind, :], axis=0)))

    return res


def smooth(idx_list, array, ele_idx_1, **kwargs):
    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    res = []

    for ind in idx_list:
        res.append(np.var(np.mean(array[ind, :], axis=0)))

    return res


def svd(idx_list, array, ele_idx_1, **kwargs):
    """
    Determines condition number for given zap indices in idx_list.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : np.ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements.
    ele_idx_1 : np.ndarray of float [n_ele]
        Element indices for which the optimization is performed.

    Returns
    -------
    res : np.ndarray of float [n_combs]
        Condition number. Lower values indicate more orthogonal e-field combinations (better).
    """
    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    res = []

    for ind in idx_list:
        s = scipy.linalg.svd(array[ind, :], compute_uv=False)
        res.append(np.max(s) / np.min(s))

    return res


def variability(idx_list, array, ele_idx_1, **kwargs):
    """
    Determines variability score for given zap indices in idx_list.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : np.ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements.
    ele_idx_1 : np.ndarray of float [n_ele]
        Element indices for which the optimization is performed.

    Returns
    -------
    res : np.ndarray of float [n_combs]
        Condition number. Lower values indicate more orthogonal e-field combinations (better).
    """
    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    res = []

    for ind in idx_list:
        distances = np.zeros((array.shape[1], array.shape[1]))
        d = 0

        for col_idx in range(0, array.shape[1] - 1):
            for row_idx in range(col_idx + 1, array.shape[1]):
                distances[row_idx, col_idx] = np.linalg.norm(array[ind, row_idx] - array[ind, col_idx])
                d += distances[row_idx, col_idx]

        res.append(1 / d)

    return res


def rowvec_diff_prepare(idx_list, array, ele_idx_1, **kwargs):
    """
    Computes the part of the difference matrix of row-vectors specified by the row indices in idx_list.
    Assumption: 'idx_list' must be sorted and valid within 'array'.

    :param idx_list: typing.List[int]
        List of row indices whose difference should be determined.
    :param array: numpy.typing.ArrayLike [n_coil x n_ele]
        E-field matrix of all possible coil positions.
    :param ele_idx_1: numpy.typing.ArrayLike [n_ele]
        Indices of the ROI elements that should be considered for optimization.
    :return: numpy.typing.ArrayLike [array.shape[0] x array.shape[0]] = [n_rows x n_rows]
        The difference matrix with the lenght of the difference vectors between
        pairs of row vectors specified by idx_list. All other (not calculated)
        paris of row vectors have a score of 0 in this matrix.
    """
    efields_from_coil = array
    total_num_coil_pos = array.shape[0]

    # preparation:
    # for each pair of coil configuration, determine the difference of the associated rwo vectors in the e-field matrix
    efields_diff = np.zeros((efields_from_coil.shape[0], efields_from_coil.shape[0]))
    for coil_idx_outer in idx_list:
        for coil_idx_inner in range(coil_idx_outer + 1, total_num_coil_pos):
            diff_vec = efields_from_coil[coil_idx_outer, ele_idx_1] - efields_from_coil[coil_idx_inner, ele_idx_1]
            efields_diff[coil_idx_outer, coil_idx_inner] = np.linalg.norm(diff_vec, ord=2)

    return efields_diff


def dist(idx_list, array, ele_idx_1, **kwargs):
    """
    Determines distance score for given zap indices in idx_list.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : np.ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements.
    ele_idx_1 : np.ndarray of float [n_ele]
        Element indices for which the optimization is performed.

    Returns
    -------
    res : np.ndarray of float [n_combs]
        Distance based score. Lower values indicate more equidistant sampling (better)
    """
    res = np.zeros(len(idx_list))

    array_dist = array[:, ele_idx_1]

    if array_dist.ndim == 1:
        array_dist = array_dist[:, np.newaxis]

    e_max = np.max(array_dist, axis=0)
    e_min = np.min(array_dist, axis=0)

    for j, ind in enumerate(idx_list):
        p = np.vstack((e_min, array_dist[ind, :], e_max))
        d_var = np.zeros(array_dist.shape[1])

        for i in range(array_dist.shape[1]):
            p_sort = np.sort(p[:, i])

            if p_sort[0] == p_sort[1]:
                p_sort = p_sort[1:]

            if p_sort[-2] == p_sort[-1]:
                p_sort = p_sort[:-1]

            d_var[i] = np.var(np.diff(p_sort))

        res[j] = np.mean(d_var)

    return res


def dist_svd(idx_list, array, ele_idx_1, ele_idx_2, **kwargs):
    """
    Determines distance score and condition number for given zap indices in idx_list. If c_max_idx is given,
    the distance based score is calculated only for this element.
    The condition number however is optimized for all elements in array

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : np.ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements.
    ele_idx_1 : np.ndarray of float [n_ele]
        Element indices for which the dist optimization is performed for.
    ele_idx_2 : np.ndarray of float [n_ele]
        Element indices for which the svd optimization is performed for.

    Returns
    -------
    res_dist : np.ndarray of float [n_combs]
        Distance based score. Lower values indicate more equidistant sampling (better).
    res_svd : np.ndarray of float [n_combs]
        Condition number. Lower values indicate more orthogonal e-field combinations (better).
    """
    res_dist = np.zeros(len(idx_list))
    res_svd = np.zeros(len(idx_list))

    array_dist = array[:, ele_idx_1]

    if array_dist.ndim == 1:
        array_dist = array_dist[:, np.newaxis]

    array_svd = array[:, ele_idx_2]

    if array_svd.ndim == 1:
        array_svd = array_svd[:, np.newaxis]

    e_max = np.max(array_dist, axis=0)
    e_min = np.min(array_dist, axis=0)

    for j, ind in enumerate(idx_list):

        # svd
        u, s, vh = scipy.linalg.svd(array_svd[ind, :])
        res_svd[j] = np.max(s) / np.min(s)

        # dist
        p = np.vstack((e_min, array_dist[ind, :], e_max))
        d_var = np.zeros(array_dist.shape[1])

        for i in range(array_dist.shape[1]):
            p_sort = np.sort(p[:, i])

            if p_sort[0] == p_sort[1]:
                p_sort = p_sort[1:]

            if p_sort[-2] == p_sort[-1]:
                p_sort = p_sort[:-1]

            d_var[i] = np.var(np.diff(p_sort))

        res_dist[j] = np.mean(d_var)

    return res_dist, res_svd


def dist_mc(idx_list, array, ele_idx_1, ele_idx_2, mode="cols", **kwargs):
    """
    Determines distance score and mutual coherence for given zap indices in idx_list. If c_max_idx is given,
    the distance based score is calculated only for this element.
    The condition number however is optimized for all elements in array.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : np.ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements.
    mode : str, default: "cols"
        Set if the mutual coherence is calculated w.r.t. columns or rows ("cols", "rows").
    ele_idx_1 : np.ndarray of float [n_ele]
        Element indices for which the dist optimization is performed for.
    ele_idx_2 : np.ndarray of float [n_ele]
        Element indices for which the mc optimization is performed for.

    Returns
    -------
    res_dist : np.ndarray of float [n_combs]
        Distance based score. Lower values indicate more equidistant sampling (better).
    res_mc : np.ndarray of float [n_combs]
        Mutual coherence. Lower values indicate more orthogonal e-field combinations (better).
    """
    res_dist = np.zeros(len(idx_list))
    res_mc = np.zeros(len(idx_list))

    array_dist = array[:, ele_idx_1]

    if array_dist.ndim == 1:
        array_dist = array_dist[:, np.newaxis]

    array_mc = array[:, ele_idx_2]

    if array_mc.ndim == 1:
        array_mc = array_mc[:, np.newaxis]

    e_max = np.max(array_dist, axis=0)
    e_min = np.min(array_dist, axis=0)

    for j, ind in enumerate(idx_list):

        # mc
        if mode == "cols":
            res_mc[j] = pynibs.mutual_coherence(array_mc[ind, :])
        elif mode == "rows":
            res_mc[j] = pynibs.mutual_coherence(array_mc[ind, :].transpose())
        else:
            raise NotImplementedError("Specified mode not implemented. Choose 'rows' or 'cols'.")

        # dist
        p = np.vstack((e_min, array_dist[ind, :], e_max))
        d_var = np.zeros(array_dist.shape[1])

        for i in range(array_dist.shape[1]):
            p_sort = np.sort(p[:, i])

            if p_sort[0] == p_sort[1]:
                p_sort = p_sort[1:]

            if p_sort[-2] == p_sort[-1]:
                p_sort = p_sort[:-1]

            d_var[i] = np.var(np.diff(p_sort))

        res_dist[j] = np.mean(d_var)

    return res_dist, res_mc


def coverage_prepare(idx_list, array, zap_idx, **kwargs):
    """Prepares coverage calculation.
    Determines coverage distributions for elements in idx_list given the zaps in zap_idx

    Parameters
    ----------
    idx_list : list [n_ele]
        Index lists of elements.
    array : ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements.
    zap_idx : ndarray of int
        Included zaps in coverage distribution.

    Returns
    -------
    x : ndarray of float [200 x n_ele]
        x-values of coverage distributions, defined in interval [0, 1] (element wise normalized electric field).
    y : ndarray of float [200 x n_ele]
        y-values of coverage distributions (element wise probability of already included e-fields).
    """

    n_x = 200

    x = np.zeros((n_x, len(idx_list)))
    y = np.zeros((n_x, len(idx_list)))

    kde = KernelDensity(bandwidth=0.03, kernel='gaussian')

    for j, ind in enumerate(idx_list):
        e_min = np.min(array[:, ind])
        e_max = np.max(array[:, ind])
        e_samples = (array[zap_idx, ind] - e_min) / (e_max - e_min)

        if not isinstance(e_samples, np.ndarray):
            e_samples = np.array([e_samples])

        kde_ele = kde.fit(e_samples[:, np.newaxis])
        x[:, j] = np.linspace(0, 1, n_x)
        y[:, j] = np.exp(kde_ele.score_samples(x[:, j][:, np.newaxis]))

    return x, y


def coverage(idx_list, array, x, y, ele_idx_1, **kwargs):
    """
    Determine coverage score (likelihood) for given zap indices in idx_list

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : np.ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements.
    x : np.ndarray of float [200 x n_ele]
        x-values of coverage distributions, defined in interval [0, 1] (element wise normalized electric field).
    y : np.ndarray of float [200 x n_ele]
        y-values of coverage distributions (element wise probability of already included e-fields).
    ele_idx_1 : np.ndarray of float [n_roi]
        Element indices for which the coverage optimization is performed for.

    Returns
    -------
    res : np.ndarray of float [n_combs]
        Coverage score (likelihood) for given electric field combinations. Lower values indicate that the
        new zap fills a gap which was not covered before.
    """

    p_e = np.zeros(len(idx_list))

    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    for j, ind in enumerate(idx_list):
        p_e_zap = np.zeros(array.shape[1])

        # normalized e-fields of this zap in all elements
        e_zap = (array[ind[-1], :] - np.min(array, axis=0)) / (np.max(array, axis=0) - np.min(array, axis=0))

        # determine e-field coverage probability for every element
        for i_ele in range(array.shape[1]):
            p_e_zap[i_ele] = np.interp(x=e_zap[i_ele], xp=x[:, i_ele], fp=y[:, i_ele])

        # accumulate e-field coverage over ever elements using log-likelihood
        p_e_zap[p_e_zap <= 0] = 1e-100
        p_e[j] = np.sum(p_e_zap)

    return p_e


def fim(idx_list, array, ele_idx_1, e_opt, c=None, **kwargs):
    """
    Determine difference between e-fields and optimal e-field determined using the Fisher Information Matrix.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : np.ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements.
    ele_idx_1 : np.ndarray of float [n_roi]
        Element indices for which the fim optimization is performed for.
    e_opt : np.ndarray of float [n_roi]
        Optimal electric field value(s) (target) determined by FIM method.
    c : np.ndarray of float [n_ele], optional
        Congruence factor map normalized to 1 (whole ROI) used to weight the difference between the optimal e-field
        and the candidate e-field. If None, no weighting is applied.

    Returns
    -------
    res : np.ndarray of float [n_combs]
        Difference between e-fields and optimal e-field.
    """

    if c is None:
        c = np.ones(array.shape[1])

    res = np.zeros(len(idx_list))

    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    for j, ind in enumerate(idx_list):
        res[j] = np.linalg.norm((e_opt - array[ind[-1], ele_idx_1]) * c[ele_idx_1])

    return res


def fim_svd(idx_list, array, ele_idx_1, ele_idx_2, e_opt, c=None, **kwargs):
    """
    Determine difference between e-fields and optimal e-field determined using the Fisher Information Matrix and
    condition number.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : np.ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements.
    ele_idx_1 : np.ndarray of float [n_roi_1]
        Element indices for which the fim optimization is performed for.
    ele_idx_2 : np.ndarray of float [n_roi_2]
        Element indices for which the svd optimization is performed for.
    e_opt : float
        Optimal electric field value (target) determined by FIM method.
    c : np.ndarray of float [n_ele], optional
        Congruence factor map normalized to 1 (whole ROI) used to weight the difference between the optimal e-field
        and the candidate e-field. If None, no weighting is applied.

    Returns
    -------
    res_fim : np.ndarray of float [n_combs]
        Difference between e-fields and optimal e-field.
    res_svd : np.ndarray of float [n_combs]
        Condition number. Lower values indicate more orthogonal e-field combinations (better)
    """

    if c is None:
        c = np.ones(array.shape[1])

    res_fim = np.zeros(len(idx_list))
    res_svd = np.zeros(len(idx_list))

    array_fim = array[:, ele_idx_1]

    if array_fim.ndim == 1:
        array_fim = array_fim[:, np.newaxis]

    array_svd = array[:, ele_idx_2]

    if array_svd.ndim == 1:
        array_svd = array_svd[:, np.newaxis]

    for j, ind in enumerate(idx_list):
        # fim
        intensity = np.mean(array_fim[ind[-1], :] / e_opt)
        res_fim[j] = np.linalg.norm((e_opt - intensity * array_fim[ind[-1], :]) * c[ele_idx_1])

        # svd
        u, s, vh = scipy.linalg.svd(array_svd[ind, :])
        res_svd[j] = np.max(s) / np.min(s)

    return res_fim, res_svd


def fim_mc(idx_list, array, ele_idx_1, ele_idx_2, e_opt, c=None, mode="rows", **kwargs):
    """
    Determine difference between e-fields and optimal e-field determined using the Fisher Information Matrix and
    mutual coherence.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : np.ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements.
    ele_idx_1 : np.ndarray of float [n_roi_1]
        Element indices for which the fim optimization is performed for.
    ele_idx_2 : np.ndarray of float [n_roi_2]
        Element indices for which the mc optimization is performed for.
    e_opt : float
        Optimal electric field value (target) determined by FIM method.
    c : np.ndarray of float [n_ele], optional
        Congruence factor map normalized to 1 (whole ROI) used to weight the difference between the optimal e-field
        and the candidate e-field. If None, no weighting is applied.
    mode : str

    Returns
    -------
    res_fim : np.ndarray of float [n_combs]
        Difference between e-fields and optimal e-field.
    res_mc : np.ndarray of float [n_combs]
        Mutual coherence. Lower values indicate more orthogonal e-field combinations (better)
    """

    if c is None:
        c = np.ones(array.shape[1])

    res_fim = np.zeros(len(idx_list))
    res_mc = np.zeros(len(idx_list))

    array_fim = array[:, ele_idx_1]

    if array_fim.ndim == 1:
        array_fim = array_fim[:, np.newaxis]

    array_mc = array[:, ele_idx_2]

    if array_mc.ndim == 1:
        array_mc = array_mc[:, np.newaxis]

    for j, ind in enumerate(idx_list):
        # fim
        res_fim[j] = np.linalg.norm((e_opt - array_fim[ind[-1], :]) * c[ele_idx_1])

        # mc
        if mode == "cols":
            res_mc[j] = pynibs.mutual_coherence(array_mc[ind, :])
        elif mode == "rows":
            res_mc[j] = pynibs.mutual_coherence(array_mc[ind, :].transpose())
        else:
            raise NotImplementedError("Specified mode not implemented. Choose 'rows' or 'cols'.")

    return res_fim, res_mc


def rowvec_diff(candidate_coil_idcs, selected_coil_idcs, efields_diff_mat):
    """
    Given a difference matrix (e.g. of row vectors/coil configurations) this function
    returns the coil configuration out of all available configurations exhibiting the
    highest minimum difference to the already selected configurations.

    Parameters
    ----------
    candidate_coil_idcs : np.ndarry[int]
        List of indices of coil configurations that are still available to pick for the optiized sequence.
    selected_coil_idcs : np.ndarray[int]
        List of indices of coil configurations that have already been selected for the optimized sequence.
    efields_diff_mat : np.ndarray[float], [n_coil,n_coil]
        Difference matrix, where each cell denotes the magnitude of the difference vector between
        two coil configurations (determined by row_idx,col_idx).

    Returns
    -------
    coil_idx: int
        Index of coil configuration with maximal minimal difference to the set of already selected coil configurations.
    """

    # min_diff_selected_to_all_coil_pos = matrix with:
    #   rows -> set of selected coil configurations (selected_coil_idcs)
    #   columns -> available coil configuration for optimization (candidate_coil_idcs)
    min_diff_selected_to_all_coil_pos = np.min(efields_diff_mat[selected_coil_idcs][:, candidate_coil_idcs], axis=0)

    # returned index valid in the "idx_list" array
    return np.argmax(min_diff_selected_to_all_coil_pos), np.max(min_diff_selected_to_all_coil_pos)



import os
import copy
import pynibs
import numpy as np
import _pickle as pickle
import multiprocessing.pool
from _functools import partial
from scipy.interpolate import LinearNDInterpolator


def workhorse_interp(idx_list, interp, params):
    """
    Single core workhorse to interpolate data.

    Parameters
    ----------
    idx_list : np.nfarray or list of float
        (n_interpolations) Indices in params array where the interpolation has to be performed
        (subset of all indices in params array).
    interp : instance of scipy.interpolate
        Interpolator instance.
    params : np.ndarray of float
        (N_interpolations, N_params) Array containing the parameters the function is evaluated
        (total array with all parameters).

    Returns
    -------
    res : np.ndarray of float
        (n_interpolations) Interpolation results (subset params[idx_list, :]).
    """
    return interp(params[idx_list, 0], params[idx_list, 1], params[idx_list, 2]) / 2.2  # 2.2


def load_cell_model(fn_csv):
    """
    Load interpolation points of the mean field model from the specified CSV file.

    Parameters
    ----------
    fn_csv : str
        Fully qualified path to the CSV containing the interpolation points of the mean field model.

    Returns
    -------
    scipy.interpolate.LinearNDInterpolator
    interpolation points 'theta'
    interpolation points 'gradient'
    """
    cell_simulation_data = [np.genfromtxt(fn_csv, delimiter=',')]

    thresholds = cell_simulation_data[-1][:, 2]
    theta = cell_simulation_data[-1][:, 1]
    rel_grad = cell_simulation_data[-1][:, 0]

    return LinearNDInterpolator(list(zip(theta, rel_grad)), thresholds), theta, rel_grad


# TODO: implement the creation of a response interpolator
def _create_model_response_interpolator(fn_model_csv):
    return LinearNDInterpolator([(0, 0), (0, 1), (1, 0), (1, 1)], [1, 1, 1, 1], fill_value=1)


def calc_e_threshold(layerid, theta, gradient=None, mep=None, neuronmodel="sensitivity_weighting",
                     waveform="biphasic", e_thresh_subject=None):
    """
    Determine sensitivity map of electric field.

    Parameters
    ----------
    layerid : str
        Choose from the neocortical layers (e.g. "L1", "L23", "L4", "L5", "L6") to load data for.
    theta : np.ndarray
        (N_stim, N_ele) Theta angle (matrix) of electric field with respect to surface normal. In degrees [0 .. 180].
    gradient : np.ndarray, optional
        (N_stim, N_ele) Electric field gradient (matrix) between layer 1 and layer 6. Optional, the neuron mean field
        model is more accurate when provided. Percent [-100 .. 100].
    mep : np.ndarray of float, optional
        (N_stim, ) MEP data (required in case of "IOcurve" approach (neuronmodel)
    neuronmodel : str, default: 'sensitivity_weighting'
        Select neuron model to modify the electric field values

        - "sensitivity_weighting": normalize threshold map and divide raw e-field by it.
        - "threshold_subtract": subtract mean threshold from electric field.
        - "threshold_binary": assign e-field a binary value to predict MEPs
            (False -> below threshold, True, above threshold).
        - "IOcurve": subtract value read from precomputed neuron IO curve from electric field.
        - "cosine" : # TODO: document
    waveform : str, default: 'biphasic'
        Waveform of TMS pulse:

        - "monophasic"
        - "biphasic"
    e_thresh_subject : float, optional
        Subject specific stimulation threshold in V/m. Typically between 60 ... 80 V/m.
        Only used for 'threshold_subtract' and '"threshold_binary",'

    Returns
    -------
    e_sens : np.ndarray
        (N_stim, N_ele) Electric field sensitivity maps.
    """
    # load neuron models
    ####################################################################################################################
    models_folder = os.path.join(pynibs.__datadir__, "neuron", "models")
    interp_folder = os.path.join(pynibs.__datadir__, "neuron", "interpolators")
    scaling_factor = None

    if waveform == "monophasic":
        models = {
            "L23": os.path.join(models_folder, "L23_PC_cADpyr_monophasic_v1.csv"),
            "L4SBC": os.path.join(models_folder, "L4_SBC_monophasic_v1.csv"),
            "L4NBC": os.path.join(models_folder, "L4_NBC_monophasic_v1.csv"),
            "L4LBC": os.path.join(models_folder, "L4_LBC_monophasic_v1.csv"),
            "L5": os.path.join(models_folder, "L5_TTPC2_cADpyr_monophasic_v1.csv")
        }

        models_io = {
            "L23": os.path.join(interp_folder, "L23_biphasic_recruitment_rate_interpolator_inverse.pkl"),
            "L5": os.path.join(interp_folder, "L5_biphasic_recruitment_rate_interpolator_inverse.pkl")
        }
    elif waveform == "biphasic":
        models = {
            "L23": os.path.join(models_folder, "L23_PC_cADpyr_biphasic_v1.csv"),
            "L4SBC": os.path.join(models_folder, "L4_SBC_biphasic_v1.csv"),
            "L4NBC": os.path.join(models_folder, "L4_NBC_biphasic_v1.csv"),
            "L4LBC": os.path.join(models_folder, "L4_LBC_biphasic_v1.csv"),
            "L5": os.path.join(models_folder, "L5_TTPC2_cADpyr_biphasic_v1.csv")
        }

        models_io = {
            "L23": os.path.join(interp_folder, "L23_biphasic_recruitment_rate_interpolator_inverse.pkl"),
            "L5": os.path.join(interp_folder, "L5_biphasic_recruitment_rate_interpolator_inverse.pkl")
        }
    else:
        raise NotImplementedError(f"Specified waveform {waveform} not implemented.")

    if neuronmodel in ["threshold_subtract", "threshold_binary", "sensitivity_weighting", "cosine"]:
        interp, thetas, rel_gradients = load_cell_model(models[layerid])

        if neuronmodel in ["sensitivity_weighting", "cosine"]:
            scaling_factor = interp(0, 0)
            if e_thresh_subject is not None:
                print(f"e_thresh_subject={e_thresh_subject} is not used for neuronmodel={neuronmodel}")

        elif e_thresh_subject is not None:
            scaling_factor = np.mean(interp(np.linspace(0, 180, 181),
                                            np.zeros(181))) / e_thresh_subject

        else:
            # Scaling factor between electric field thresholds of model and subject specific e-field thresholds.
            # scaling_factor = e_threshold_model / e_threshold_subject
            # Model thresholds are higher than subject specific thresholds (typically in a range between 2..3)
            # average thresholds between 0 and 180° at E_grad = 0
            scaling_factor = 1

    elif neuronmodel == "IOcurve":
        _, thetas, rel_gradients = load_cell_model(models[layerid])

        # TODO: not implemented yet
        if not os.path.exists(models_io[layerid]):
            raise NotImplementedError("[neuron_regression] Pickl files containing the response interpolators "
                                      f"do not exist (path checked: {models_io[layerid]}) and their creation "
                                      "is not implemented yet.")
            # interp = _create_model_response_interpolator(models[layerid])
            # with open(models_io[layerid], 'wb') as f:
            #    pickle.dump(interp, f)
        else:
            with open(models_io[layerid], 'rb') as f:
                interp = pickle.load(f)

    else:
        raise NotImplementedError(f"Specified neuronmodel {neuronmodel} not implemented.")

    # bound observed values to min/max values available in the model
    ####################################################################################################################
    theta_bound = theta
    theta_bound[np.where(theta > np.max(thetas))] = np.max(thetas)
    theta_bound[np.where(theta < np.min(thetas))] = np.min(thetas)

    if gradient is None:
        gradient_bound = np.zeros(theta.shape)
    else:
        gradient_bound = gradient
        gradient_bound[np.where(gradient > np.max(rel_gradients))] = np.max(rel_gradients)
        gradient_bound[np.where(gradient < np.min(rel_gradients))] = np.min(rel_gradients)

    # Determine approach specific effective electric field
    ####################################################################################################################
    if neuronmodel in ["threshold_subtract", "threshold_binary", "sensitivity_weighting"]:
        e_thres = interp(theta_bound, gradient_bound) / scaling_factor

    elif neuronmodel == "cosine":
        e_thres = scaling_factor * 1 / np.abs(np.cos(theta_bound / np.pi))

    elif neuronmodel == "IOcurve":
        # normalize MEPs between [0, 0.999]
        mep_threshold = 2
        mep_cropped = copy.deepcopy(mep)
        mep_cropped[mep > mep_threshold] = mep_threshold
        mep_norm = mep_cropped / (mep_threshold * 1.05)

        # calculate expected electric field at observed MEP
        params = np.zeros((theta.shape[0] * theta.shape[1], 3))
        params[:, 0] = gradient_bound.flatten()
        params[:, 1] = theta_bound.flatten()
        params[:, 2] = np.repeat(mep_norm, theta.shape[1])

        idx = np.arange(params.shape[0])
        idx_chunked = pynibs.compute_chunks(list(idx), multiprocessing.cpu_count())

        pool = multiprocessing.Pool(multiprocessing.cpu_count())
        workhorse_partial = partial(workhorse_interp, interp=interp, params=params)
        res = np.hstack(pool.map(workhorse_partial, idx_chunked))
        e_thres = np.reshape(res, theta.shape)
        pool.close()
        pool.join()

    else:
        raise NotImplementedError

    return e_thres


def calc_e_effective(e, layerid, theta, gradient=None, neuronmodel="sensitivity_weighting", mep=None,
                     waveform="biphasic", e_thresh_subject=None):
    """
    Determines the effective electric field using a neuron mean field model.
    The electric field magnitude is 'subtracted' by the threshold map (in V/m), yielding the
    effective electric field (e_eff).

    Parameters
    ----------
    e : np.ndarray
        (N_stim, N_ele) Electric field (matrix).
    layerid : str
        Choose from the neocortical layers (e.g. "L1", "L23", "L4", "L5", "L6").
    theta : np.ndarray
        (N_stim, N_ele) Theta angle (matrix)  of electric field with respect to surface normal.
    gradient : np.ndarray, optional
        (N_stim, N_ele) Electric field gradient (matrix)  between layer 1 and layer 6. Optional, the neuron mean field
        model is more accurate when provided.
    neuronmodel : str, default: 'threshold'
        Select neuron model to modify the electric field values

        - "sensitivity_weighting": normalize threshold map and divide raw e-field by it.
        - "threshold_subtract": subtract mean threshold from electric field.
        - "threshold_binary": assign e-field a binary value to predict MEPs
          (False -> below threshold, True, above threshold).
        - "IOcurve": subtract value read from precomputed neuron IO curve from electric field.
    mep : np.ndarray of float [N_stim], optional
        MEP data (required in case of "IOcurve" approach (neuronmodel)).
    waveform : str, default: 'biphasic'
        Waveform of TMS pulse:

        - "monophasic"
        - "biphasic"
    e_thresh_subject : float, optional
        Subject specific stimulation threshold in V/m. Typically, between 60 ... 80 V/m.
        This is not used for sensitivity_weighting and

    Returns
    -------
    e_eff : np.ndarray
        Effective electric field (matrix) [N_stim x N_ele] the regression analysis can be performed with.
    """
    # determine sensitivity map
    e_thres = calc_e_threshold(layerid=layerid,
                               theta=theta,
                               gradient=gradient,
                               neuronmodel=neuronmodel,
                               mep=mep,
                               waveform=waveform,
                               e_thresh_subject=e_thresh_subject)

    if neuronmodel == "threshold_subtract":
        e_eff = e - e_thres
    elif neuronmodel == "threshold_binary":
        e_eff = e > e_thres
    elif neuronmodel in ["sensitivity_weighting", "cosine"]:
        e_eff = e / e_thres
    else:
        e_eff = np.zeros(e.shape)

    return e_eff



import numpy as np

import pynibs


def stimulation_threshold(elm_idx_list, mep, mep_params, n_samples, e, c_factor_percentile=95, mep_threshold=0.5,
                          c_factor=None, c_function=None, t_function=None):
    """
    Computes the stimulation threshold in terms of the electric field in [V/m]. The threshold is defined as the
    electric field value where the mep exceeds mep_threshold. The average value is taken over all mep curves in each
    condition and over an area where the congruence factor exceeds ``c_factor_percentile``.

    Parameters
    ----------
    elm_idx_list : np.ndarray
        (chunksize) List of element indices, the congruence factor is computed for.
    mep : list of Mep object instances
        (n_cond) List of fitted :py:class:`~pynibs.expio.Mep` object instances for all conditions.
    mep_params : np.ndarray of float [n_mep_params_total]
        List of all mep parameters of curve fits used to calculate the MEP (accumulated into one array)

        * e.g. [``mep_#1_para_#1``, ``mep_#1_para_#2``, ``mep_#1_para_#3``, ``mep_#2_para_#1``, ``mep_#2_para_#1``, ...]

    n_samples : int
        Number of data points to generate discrete mep and e curves.
    e : list of list of np.ndarray of float
        [n_cond][n_datasets][n_elm] Tuple of ``n_datasets`` of the electric field to compute the congruence factor for,
        e.g. ``(e_mag, e_norm, e_tan)``.
        Each dataset is a list over all conditions containing the electric field component of interest

        * e.g.: ``len(e) = n_cond``
        * ``len(e[0]) = n_comp`` (e.g: ``e_mag = e[0])``)
    c_factor_percentile : float
        Percentile of the c_factor taken into account for the threshold evaluation. Only c_factors are considered
        exceeding this.
    mep_threshold : float
        MEP value in [mV], which has to be exceeded for threshold definition.
    c_factor : np.ndarray of float
    (n_roi, n_datasets) Congruence factor in each element specified in elm_idx_list and for each input dataset.
    c_function : function
        Defines the function to use during c_gpc to calculate the congruence factor.

        * congruence_factor_curveshift_workhorse: determines the average curve shift
        * congruence_factor_curveshift_workhorse_stretch_correction: determines the average curve shift
        * congruence_factor_curveshift_workhorse_stretch_correction_variance: determines the average curve shift
        * congruence_factor_variance_workhorse: evaluates the variance of the shifting and stretching parameters

    t_function : function
        Defines the function to determine the stimulation_threshold.

        * stimulation_threshold_mean_mep_threshold: uses mep_threshold to determine the corresponding e_threshold over
          all conditions and takes the average values as the stimulation threshold
        * stimulation_threshold_pynibs.sigmoid: Fits a new pynibs.sigmoid using all datapoints in the mep-vs-E space and
          evaluates the threshold from the turning point or the intersection of the derivative in the crossing point
          with the e-axis

    Returns
    -------
    stim_threshold_avg: float
        Average stimulation threshold in [V/m] where c_factor is greater than ``c_factor_percentile``.
    """

    n_datasets = len(e[0])
    n_conditions = len(mep)
    mep_params = np.array(mep_params).flatten()

    # rearrange mep parameters to individual conditions
    mep_params_cond = []
    start_idx = 0

    for i_cond in range(n_conditions):
        mep_params_cond.append(mep_params[start_idx:(start_idx + mep[i_cond].popt.size)])
        start_idx = start_idx + mep[i_cond].popt.size

    # calculate mep curves per condition
    mep_curve = []
    intensities = []

    for i_cond in range(n_conditions):
        intensities.append(np.linspace(mep[i_cond].x_limits[0], mep[i_cond].x_limits[1], n_samples))
        mep_curve.append(mep[i_cond].eval(intensities[-1], mep_params_cond[i_cond]))

    # determine congruence factor, if not provided
    if not c_factor.any():
        if c_function == pynibs.congruence.cf_curveshift_workhorse or \
                c_function == pynibs.congruence.cf_curveshift_workhorse_stretch_correction or \
                c_function == pynibs.congruence.cf_curveshift_workhorse_stretch_correction_variance:
            c_factor = c_function(elm_idx_list,
                                  mep=mep,
                                  mep_params=mep_params,
                                  n_samples=n_samples,
                                  e=e)

        elif c_function == pynibs.congruence.cf_variance_workhorse:
            c_factor = c_function(elm_idx_list,
                                  mep=mep,
                                  mep_params=mep_params,
                                  e=e)

    # determine elements where the congruence factor exceeds c_factor_percentile
    elm_idx = []
    c_factor_percentile_value = []

    for i_data in range(n_datasets):
        c_factor_percentile_value.append(np.percentile(c_factor[np.logical_not(np.isnan(c_factor[:, i_data])), i_data],
                                                       c_factor_percentile))
        elm_idx.append(np.where(c_factor[:, i_data] > c_factor_percentile_value[i_data])[0])

    if t_function == mean_mep_threshold:
        stim_threshold_avg, stim_threshold_std = \
            mean_mep_threshold(elm_idx=elm_idx,
                               mep_curve=mep_curve,
                               intensities=intensities,
                               e=e,
                               mep_threshold=mep_threshold)

    elif t_function == sigmoid_thresh:
        stim_threshold_avg, stim_threshold_std = \
            sigmoid_thresh(elm_idx=elm_idx,
                           mep_curve=mep_curve,
                           intensities=intensities,
                           e=e,
                           mep_threshold=mep_threshold)

    elif t_function == intensity_thresh:
        stim_threshold_avg = \
            intensity_thresh(mep_curve=mep_curve,
                             intensities=intensities,
                             mep_threshold=mep_threshold)
        stim_threshold_std = np.nan

    else:
        raise NotImplementedError('Provided t_function not implemented yet!')

    return stim_threshold_avg, stim_threshold_std


def mean_mep_threshold(elm_idx, mep_curve, intensities, e, mep_threshold):
    """
    Determines the stimulation threshold by calculating the average electric field over all conditions, where the
    mep curves exceed the value of mep_threshold (in [mV]).

    Parameters
    ----------
    elm_idx : list of np.ndarray of int
        [n_datasets](n_elements) Element indices where the congruence factor exceeds a certain percentile,
        defined during the call of :py:meth:`stimulation_threshold`.
    mep_curve : list  of np.ndarray of float
        [n_conditions](n_samples) MEP curve values for every condition.
    intensities : list  of np.ndarray of float
        [n_conditions](n_samples) To the MEP values corresponding stimulator intensities in [A/us].
    e : list of list of np.ndarray of float
        [n_cond][n_datasets][n_elm] Tuple of n_datasets of the electric field to compute the congruence factor for,
        e.g. ``(e_mag, e_norm, e_tan)``.
        Each dataset is a list over all conditions containing the electric field component of interest

        * e.g.: ``len(e) = n_cond``
        * ``len(e[0]) = n_comp`` (e.g: ``e_mag = e[0])``)

    mep_threshold : float
        MEP value in [mV], which has to be exceeded for threshold definition.

    Returns
    -------
    stim_threshold_avg : float
        Average stimulation threshold in [V/m] where c_factor is greater than c_factor_percentile
    """

    n_conditions = len(mep_curve)
    n_datasets = len(e[0])

    # determine electric field values exceeding mep_threshold in this elements
    stim_threshold_cond = [np.zeros((elm_idx[i_data].size, n_conditions)) * np.nan for i_data in range(n_datasets)]
    stim_threshold_avg = [np.nan for _ in range(n_datasets)]
    stim_threshold_std = [np.nan for _ in range(n_datasets)]

    for i_cond in range(n_conditions):
        e_threshold_idx = np.where(mep_curve[i_cond] > mep_threshold)[0]

        if e_threshold_idx.any():
            for i_data in range(n_datasets):
                stim_threshold_cond[i_data][:, i_cond] = e[i_cond][i_data][elm_idx[i_data]] * \
                                                         intensities[i_cond][e_threshold_idx[0]]

    for i_data in range(n_datasets):
        stim_threshold_avg[i_data] = np.mean(
                stim_threshold_cond[i_data][np.logical_not(np.isnan(stim_threshold_cond[i_data]))])
        stim_threshold_std[i_data] = np.std(
                stim_threshold_cond[i_data][np.logical_not(np.isnan(stim_threshold_cond[i_data]))])

    return stim_threshold_avg, stim_threshold_std


def sigmoid_thresh(elm_idx, mep_curve, intensities, e, mep_threshold):
    """
    Determines the stimulation threshold by calculating an equivalent :py:class:`pynibs.expio.Mep.sigmoid`
    over all conditions. The stimulation threshold is the electric field value where the mep curves exceed the value of
    mep_threshold (in [mV]).

    Parameters
    ----------
    elm_idx : list of np.ndarray of int
        [n_datasets](n_elements) Element indices where the congruence factor exceeds a certain percentile,
        defined during the call of :py:meth:`stimulation_threshold`.
    mep_curve : list  of np.ndarray of float
        [n_conditions](n_samples) MEP curve values for every condition.
    intensities : list  of np.ndarray of float
        [n_conditions](n_samples) To the MEP values corresponding stimulator intensities in [A/us].
    e : list of list of np.ndarray of float
        [n_cond][n_datasets][n_elm] Tuple of n_datasets of the electric field to compute the congruence factor for,
        e.g. ``(e_mag, e_norm, e_tan)``.
        Each dataset is a list over all conditions containing the electric field component of interest

        * e.g.: ``len(e) = n_cond``
        * ``len(e[0]) = n_comp`` (e.g: ``e_mag = e[0])``)

    mep_threshold : float
        MEP value in [mV], which has to be exceeded for threshold definition

    Returns
    -------
    stim_threshold_avg : float
        Average stimulation threshold in [V/m] where c_factor is greater than c_factor_percentile
    """

    n_conditions = len(mep_curve)
    n_datasets = len(e[0])
    stim_threshold_elm = [[] for _ in range(n_datasets)]
    stim_threshold_avg = [[] for _ in range(n_datasets)]
    stim_threshold_std = [[] for _ in range(n_datasets)]

    # accumulate all data values in one array
    mep_curve_all = np.hstack(mep_curve)

    for i_data in range(n_datasets):
        print(('Evaluating stimulation threshold for dataset {}/{}'.format(i_data + 1, n_datasets)))
        n_elms = len(elm_idx[i_data])
        stim_threshold_elm[i_data] = np.zeros(n_elms) * np.nan

        for i_elm, elm in enumerate(elm_idx[i_data]):
            print((' > Element {}/{}'.format(i_elm, n_elms)))

            # accumulate all data values in one array
            e_all = []

            for i_cond in range(n_conditions):
                e_all.append(e[i_cond][i_data][elm] * intensities[i_cond])
            e_all = np.hstack(e_all)

            # fit data to function
            mep = pynibs.Mep(intensities=e_all, mep=mep_curve_all, intensity_min_threshold=0, mep_min_threshold=0)
            mep.fit = mep.run_fit_multistart(pynibs.expio.fit_funs.sigmoid,
                                             x=e_all,
                                             y=mep_curve_all,
                                             p0=[70, 0.6, 2],
                                             constraints=None,
                                             verbose=False,
                                             n_multistart=20)

            # read out optimal function parameters from best fit
            try:
                for p in ['x0', 'r', 'amp']:
                    mep.popt.append(mep.fit.best_values[p])

                mep.popt = np.asarray(mep.popt)
                mep.cvar = np.asarray(mep.fit.covar)
                mep.pstd = np.sqrt(np.diag(mep.cvar))
                mep.fun = pynibs.expio.fit_funs.sigmoid

                # determine stimulation threshold
                e_fit = np.linspace(np.min(e_all), np.max(e_all), 200)
                mep_fit = mep.eval_opt(e_fit)
                e_threshold_idx = np.where(mep_fit > mep_threshold)[0]

                if e_threshold_idx.any():
                    stim_threshold_elm[i_data][i_elm] = e_fit[e_threshold_idx[0]]

            except (AttributeError, ValueError):
                print(' > Warning: pynibs.sigmoid in element could not be fitted!')
                stim_threshold_elm[i_data][i_elm] = np.nan

        # determine mean threshold over all elements
        stim_threshold_avg[i_data] = np.mean(stim_threshold_elm[i_data]
                                             [np.logical_not(np.isnan(stim_threshold_elm[i_data]))])
        stim_threshold_std[i_data] = np.std(stim_threshold_elm[i_data]
                                            [np.logical_not(np.isnan(stim_threshold_elm[i_data]))])

    return stim_threshold_avg, stim_threshold_std


def intensity_thresh(mep_curve, intensities, mep_threshold):
    """
    Determines the stimulation threshold of one particular condition (usually the most sensitive e.g. M1-45). The
    stimulation threshold is the stimulator intensity value in [A/us] where the mep curves exceed the value of
    mep_threshold (in [mV]).

    Parameters
    ----------
    mep_curve: list [1] of np.ndarray of float [n_samples]
        MEP curve values for every conditions
    intensities: list [1] of np.ndarray of float [n_samples]
        To the MEP values corresponding stimulator intensities in [A/us]
    mep_threshold: float
        MEP value in [mV], which has to be exceeded for threshold definition

    Returns
    -------
    stim_threshold_avg: float
        Average stimulation threshold in [V/m] where c_factor is greater than c_factor_percentile
    """

    stim_threshold = np.nan
    i_threshold_idx = np.where(mep_curve[0] > mep_threshold)[0]

    if i_threshold_idx.any():
        stim_threshold = intensities[0][i_threshold_idx[0]]

    return stim_threshold



"""
This holds methods for the TMS-based cortical localization approach as published in [1]_

References
----------
.. [1] Numssen, O., Zier, A. L., Thielscher, A., Hartwigsen, G., Knösche, T. R., & Weise, K. (2021).
   Efficient high-resolution TMS mapping of the human motor cortex by nonlinear regression. NeuroImage, 245, 118654.
"""
from .regression import *
from .score_types import *
from .dual_node_detection import *



import gdist
import pynibs
import unittest
import numpy as np


class TestGeodesicDist(unittest.TestCase):
    """
    Test pynibs.quality_measures.geodesic_dist()
    """
    nodes1 = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
    tris1 = np.array([[0, 1, 2]])
    nodes2 = np.array([[0, 0, 0], [1, 1, 1], [5, 2, 0], [5, 3, 5]])
    tris2 = np.array([[0, 1, 2], [1, 2, 3]])
    source_node = np.array([0])
    source_tri = 1
    source_is_node_node = True
    source_is_node_tri = False

    def test_geodesic_dist_node(self):
        # Test case 1: source is a node
        nodes_dist, tris_dist = pynibs.geodesic_dist(self.nodes1, self.tris1, self.source_node,
                                                     self.source_is_node_node)

        assert isinstance(nodes_dist, np.ndarray)
        assert isinstance(tris_dist, np.ndarray)
        assert np.allclose(nodes_dist, [0, np.sqrt(3), 2 * np.sqrt(3)])
        assert np.allclose(tris_dist, [np.sqrt(3)])

    '''
    # doesn't work yet:
    def test_geodesic_dist_tri(self):  
        # Test case 2: source is a triangle
        nodes_dist, tris_dist = pynibs.geodesic_dist(self.nodes2, self.tris2, self.source_tri, self.source_is_node_tri)

        assert isinstance(nodes_dist, np.ndarray)
        assert isinstance(tris_dist, np.ndarray)
        assert np.allclose(nodes_dist, ?)
        assert np.allclose(tris_dist, ?)
    '''


class TestEuclideanDist(unittest.TestCase):
    """
    Test pynibs.quality_measures.euclidean_dist()
    """
    nodes = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2], [3, 3, 3]])
    tris = np.array([[0, 1, 2], [1, 2, 3]])
    source_node = np.array([0])
    source_tri = np.array([0])
    source_is_node_node = True
    source_is_node_tri = False

    def test_euclidean_dist_node(self):
        # Test case 1: source is a node
        nodes_dist, tris_dist = pynibs.euclidean_dist(self.nodes[:-1], self.tris[:-1], self.source_node,
                                                      self.source_is_node_node)

        assert isinstance(nodes_dist, np.ndarray)
        assert isinstance(tris_dist, np.ndarray)
        assert np.allclose(nodes_dist, [0, np.sqrt(3), 2 * np.sqrt(3)])
        assert np.allclose(tris_dist, [np.sqrt(3)])

    '''
    doesnt work yet:
    def test_euclidean_dist_tri(self):
        # Test case 2: source is a triangle
        nodes_dist, tris_dist = pynibs.euclidean_dist(self.nodes, self.tris, self.source_tri, self.source_is_node_tri)

        assert isinstance(nodes_dist, np.ndarray)
        assert isinstance(tris_dist, np.ndarray)
        assert np.allclose(nodes_dist, [0, np.sqrt(3), 2 * np.sqrt(3), 3 * np.sqrt(3)])
        assert np.allclose(tris_dist, [0, np.sqrt(3)])
    '''


class TestIntersectionVecPlan(unittest.TestCase):

    def test_direct_intersection(self):
        ray_dir = np.array([0, 0, -1])
        ray_origin = np.array([0, 0, 10])
        plane_n = np.array([0, 0, 1])
        plane_p = np.array([0, 0, 0])
        expected_intersection = np.array([0, 0, 0])
        intersec = pynibs.intersection_vec_plan(ray_dir, ray_origin, plane_n, plane_p)
        np.testing.assert_array_almost_equal(intersec, expected_intersection)

    def test_parallel_no_intersection(self):
        ray_dir = np.array([0, 1, 0])
        ray_origin = np.array([0, 0, 10])
        plane_n = np.array([0, 0, 1])
        plane_p = np.array([0, 0, 0])
        intersec = pynibs.intersection_vec_plan(ray_dir, ray_origin, plane_n, plane_p)
        assert (intersec == [np.inf, np.inf, np.inf]).all()

    def test_origin_within_plane(self):
        ray_dir = np.array([0, 0, 1])
        ray_origin = np.array([0, 0, 0])
        plane_n = np.array([0, 0, 1])
        plane_p = np.array([0, 0, 0])
        expected_intersection = np.array([0, 0, 0])
        intersec = pynibs.intersection_vec_plan(ray_dir, ray_origin, plane_n, plane_p)
        np.testing.assert_array_almost_equal(intersec, expected_intersection)

    def test_edge_case_near_parallel(self):
        ray_dir = np.array([0.001, 0.001, -1])
        ray_origin = np.array([0, 0, 10])
        plane_n = np.array([0, 0, 1])
        plane_p = np.array([0, 0, 0])
        # Expected intersection is approximate due to near parallel direction
        expected_intersection = np.array([0.01, 0.01, 0])
        intersec = pynibs.intersection_vec_plan(ray_dir, ray_origin, plane_n, plane_p)
        np.testing.assert_array_almost_equal(intersec, expected_intersection, decimal=2)


if __name__ == '__main__':
    unittest.main()



import os
import meshio
import trimesh
import warnings
import nibabel
import numpy as np
from tqdm import tqdm
from scipy.interpolate import griddata
from vtkmodules.util import numpy_support  # don't import vtk due to opengl issues when running headless
from vtkmodules.util.vtkConstants import VTK_TRIANGLE
from vtkmodules.vtkIOXML import vtkXMLUnstructuredGridWriter
from vtkmodules.vtkCommonDataModel import vtkUnstructuredGrid
from vtkmodules.vtkCommonCore import vtkPoints, vtkDoubleArray
from vtkmodules.vtkFiltersCore import vtkCellDataToPointData, vtkPointDataToCellData

import pynibs


def cell_data_to_point_data(tris, data_tris, nodes, method='nearest'):
    """
    A wrapper for scipy.interpolate.griddata to interpolate cell data to node data.

    Parameters
    ----------
    tris : np.ndarray
        (n_tri, 3) element number list.
    data_tris : np.ndarray
        (n_tri x 3) data in tris.
    nodes : np.ndarray
        (n_nodes, 3) nodes coordinates.
    method: str, default: 'nearest'
        Which method to use for interpolation. Default uses NearestNDInterpolator.

    Returns
    -------
    data_nodes : np.ndarray
        Data in nodes.
    """
    elms_center = np.mean(nodes[tris], axis=1)
    return griddata(elms_center, data_tris, nodes, method)


def data_nodes2elements(data, con):
    """
    Transforms data in nodes to elements (triangles or tetrahedra).

    Parameters
    ----------
    data : np.ndarray of float
        (N_nodes, N_data) Data given in the nodes.
    con : np.ndarray of int
        triangles: (N_elements, 3).
        tetrahedra: (N_elements, 4).
        Connectivity index list forming the elements.

    Returns
    -------
    out : np.ndarray of float
        (N_elements, N_data) Data given in the element centers.
    """
    return np.average(data[con], axis=1)


def data_elements2nodes(data, con, precise=False):
    """
    Transforms data in elements (triangles or tetrahedra) to nodes.
    Data can be list of multiple data arrays.

    Parameters
    ----------
    data : np.ndarray of float or list of np.ndarray
        (N_elements, N_data) Data given in the elements (multiple datasets who fit to con may be passed in a list).
    con : np.ndarray of int
        triangles: (N_elements. 3).
        tetrahedra: (N_elements, 4).
        Connectivity index list forming the elements.
    precise : bool, default: False
        Compute data transformation precisely but slow. Better for near-0 values.

    Returns
    -------
    out : np.ndarray of float or list of np.ndarray
        (N_nodes, N_data) Data in nodes.
    """
    # check if single dataset or a list of multiple datasets is passed
    if type(data) is not list:
        single_array_input = True
        data = [data]
    else:
        single_array_input = False

    n_elements = data[0].shape[0]
    n_nodes = con.max() - con.min() + 1
    if con.min() != 0:
        warnings.warn("Node number list is not zero based")

    # built connectivity matrix
    if not precise:
        try:
            c = np.zeros([n_elements, n_nodes])

            for i in range(n_elements):
                c[i, (con[i])] = 1.0 / con.shape[1]

            # filter out NaN from dataset
            for i in range(len(data)):
                data[i][np.isnan(data[i])] = 0

            # determine inverse of node matrix
            cinv = np.linalg.pinv(c)
            # transform data from element center to element nodes
            data_in_nodes = [np.dot(cinv, d) for d in data]

            # if single array was provided, return array as well
            if single_array_input:
                data_in_nodes = np.array(data_in_nodes)

            return data_in_nodes

        except np.core._exceptions._ArrayMemoryError:
            warnings.warn("Cannot allocate enough RAM to do fast data->nodes conversion. "
                          "Falling back to (slow) iterative mapping.")
    data_in_nodes = []

    con_flat = con.flatten()
    n_dims = con.shape[1]

    for d in data:
        data_flat = np.repeat(d, n_dims)
        # data_nodes = np.zeros(n_nodes, )
        # for i in tqdm(range(n_nodes), desc="Mapping elements to node data"):
        # data_nodes[i] = d[np.argwhere(con == i)[:, 0]].mean()

        data_in_nodes.append(np.array([data_flat[con_flat == i].mean() for i in range(n_nodes)]))

    # if single array was provided, return array as well
    if single_array_input:
        data_in_nodes = np.array(data_in_nodes)
    return data_in_nodes


def project_on_scalp_hdf5(coords, mesh, scalp_tag=1005):
    """
    Find the node in the scalp closest to each coordinate.

    Parameters
    ----------
    coords: np.ndarray
        (n, 3) Vectors to be transformed.
    mesh: str or pynibs.TetrahedraLinear
        Filename of mesh in .hdf5 format or Mesh structure.
    scalp_tag: int, default: 1005
        Tag in the mesh where the scalp is to be set.

    Returns
    -------
    points_closest: np.ndarray
        (n, 3) coordinates projected scalp (closest skin points).
    """
    # read head mesh and extract skin surface
    if isinstance(mesh, str):
        mesh = pynibs.load_mesh_hdf5(mesh)

    if coords.ndim == 1:
        coords = coords[np.newaxis,]

    # crop to skin surface
    triangles_skin = mesh.triangles[mesh.triangles_regions == scalp_tag]
    point_idx_skin = np.unique(triangles_skin)
    points_skin = mesh.points[point_idx_skin]

    # find points with smalled Euclidean distance
    points_closest = np.zeros(coords.shape)
    for i, c in enumerate(coords):
        points_closest[i,] = points_skin[np.argmin(np.linalg.norm(points_skin - c, axis=1)),]

    return points_closest


def project_on_scalp(coords, mesh, scalp_tag=1005):
    """
    Find the node in the scalp closest to each coordinate

    Parameters
    ----------
    coords: nx3 np.ndarray
        Vectors to be transformed
    mesh: pynibs.TetrahedraLinear or simnibs.msh.mesh_io.Msh
        Mesh structure in simnibs or pynibs format
    scalp_tag: int, default: 1005
        Tag in the mesh where the scalp is to be set.

    Returns
    -------
    points_closest: np.ndarry
        (n, 3) coordinates projected scalp (closest skin points)
    """
    from simnibs.msh.transformations import project_on_scalp as project_on_scalp_msh
    from simnibs.msh.mesh_io import Msh

    if isinstance(mesh, pynibs.TetrahedraLinear):
        points_closest = project_on_scalp_hdf5(coords=coords, mesh=mesh, scalp_tag=scalp_tag)
    elif isinstance(mesh, Msh):
        points_closest = project_on_scalp_msh(coords=coords, mesh=mesh, scalp_tag=scalp_tag, distance=0.)
    else:
        raise ValueError(f"Unknown mesh type: {type(mesh)}.")

    return points_closest


def refine_surface(fn_surf, fn_surf_refined, center, radius, repair=True, remesh=True, verbose=True):
    """
    Refines surface (.stl) in spherical ROI and saves as .stl file.

    Parameters
    ----------
    fn_surf : str
        Input filename (.stl).
    fn_surf_refined : str
        Output filename (.stl).
    center : np.ndarray of float
        (3) Center of spherical ROI (x,y,z).
    radius : float
        Radius of ROI.
    repair : bool, default: True
        Repair surface mesh to ensure that it is watertight and forms a volume.
    remesh : bool, default: False
        Perform remeshing with meshfix (also removes possibly overlapping facets and intersections).
    verbose : bool, default: True
        Print output messages.

    Returns
    -------
    <file>: .stl file
    """
    radius_ = radius + 2
    refine = True

    while refine:
        if verbose:
            print(f"Loading {fn_surf} ...")
        # reading original .stl file
        wm = trimesh.load(fn_surf)

        tris = wm.faces
        tris_center = wm.triangles_center
        points = wm.vertices

        # Splitting elements by adding tris_center to points in ROI
        mask_roi = np.linalg.norm(tris_center - center, axis=1) < radius
        ele_idx_roi = np.where(np.linalg.norm(tris_center - center, axis=1) < radius)[0]
        points_refine = points
        tris_refine = tris

        if verbose:
            print(f"Splitting elements ...")

        for ele_idx in tqdm(ele_idx_roi):
            points_idx_ele = tris[ele_idx, :]
            p_0 = points[points_idx_ele[0], :]
            p_1 = points[points_idx_ele[1], :]
            p_2 = points[points_idx_ele[2], :]
            p_01 = p_0 + 0.5 * (p_1 - p_0)
            p_02 = p_0 + 0.5 * (p_2 - p_0)
            p_12 = p_1 + 0.5 * (p_2 - p_1)

            points_refine = np.vstack((points_refine, p_01, p_02, p_12))

            mask_roi = np.hstack((mask_roi, False, False, False, False))

            # add 6 new triangles
            p_0_idx = points_idx_ele[0]
            p_1_idx = points_idx_ele[1]
            p_2_idx = points_idx_ele[2]
            p_01_idx = points_refine.shape[0] - 3
            p_02_idx = points_refine.shape[0] - 2
            p_12_idx = points_refine.shape[0] - 1

            # adding 4 elements
            tris_refine = np.vstack((tris_refine, np.array([[p_0_idx, p_01_idx, p_02_idx],
                                                            [p_01_idx, p_1_idx, p_12_idx],
                                                            [p_02_idx, p_12_idx, p_2_idx],
                                                            [p_01_idx, p_12_idx, p_02_idx]])))

        ele_idx_del = []

        if radius != np.inf:
            if verbose:
                print(f"Adding triangles in surrounding elements ...")
            # add triangles in surrounding elements
            ele_sur_idx = np.where(np.logical_and(np.linalg.norm(tris_center - center, axis=1) < radius_,
                                                  np.linalg.norm(tris_center - center, axis=1) >= radius))[0]

            for ele_sur in tqdm(ele_sur_idx):
                points_idx_ele = tris[ele_sur, :]
                p_0 = points[points_idx_ele[0], :]
                p_1 = points[points_idx_ele[1], :]
                p_2 = points[points_idx_ele[2], :]
                p_01 = p_0 + 0.5 * (p_1 - p_0)
                p_02 = p_0 + 0.5 * (p_2 - p_0)
                p_12 = p_1 + 0.5 * (p_2 - p_1)

                p_0_idx = points_idx_ele[0]
                p_1_idx = points_idx_ele[1]
                p_2_idx = points_idx_ele[2]

                p_on_02 = False
                p_on_12 = False
                p_on_01 = False

                if (np.sum(p_01 == points_refine, axis=1) == 3).any():
                    p_on_01 = True

                if (np.sum(p_02 == points_refine, axis=1) == 3).any():
                    p_on_02 = True

                if (np.sum(p_12 == points_refine, axis=1) == 3).any():
                    p_on_12 = True

                # no edge with point
                if not p_on_01 and not p_on_02 and not p_on_12:
                    pass

                # one edge with point
                elif p_on_01 and not p_on_02 and not p_on_12:
                    ele_idx_del.append(ele_sur)
                    p_01_idx = np.where(np.sum(points_refine == p_01, axis=1) == 3)[0][0]
                    tris_refine = np.vstack((tris_refine, np.array([[p_0_idx, p_01_idx, p_2_idx],
                                                                    [p_01_idx, p_1_idx, p_2_idx]])))

                elif p_on_02 and not p_on_01 and not p_on_12:
                    ele_idx_del.append(ele_sur)
                    p_02_idx = np.where(np.sum(points_refine == p_02, axis=1) == 3)[0][0]
                    tris_refine = np.vstack((tris_refine, np.array([[p_0_idx, p_1_idx, p_02_idx],
                                                                    [p_02_idx, p_1_idx, p_2_idx]])))

                elif p_on_12 and not p_on_02 and not p_on_01:
                    ele_idx_del.append(ele_sur)
                    p_12_idx = np.where(np.sum(points_refine == p_12, axis=1) == 3)[0][0]
                    tris_refine = np.vstack((tris_refine, np.array([[p_0_idx, p_1_idx, p_12_idx],
                                                                    [p_0_idx, p_12_idx, p_2_idx]])))

                # 2 edges with points
                elif p_on_02 and p_on_12 and not p_on_01:
                    ele_idx_del.append(ele_sur)
                    p_12_idx = np.where(np.sum(points_refine == p_12, axis=1) == 3)[0][0]
                    p_02_idx = np.where(np.sum(points_refine == p_02, axis=1) == 3)[0][0]
                    tris_refine = np.vstack((tris_refine, np.array([[p_0_idx, p_1_idx, p_02_idx],
                                                                    [p_1_idx, p_12_idx, p_02_idx],
                                                                    [p_02_idx, p_12_idx, p_2_idx]])))

                elif p_on_02 and p_on_01 and not p_on_12:
                    ele_idx_del.append(ele_sur)
                    p_01_idx = np.where(np.sum(points_refine == p_01, axis=1) == 3)[0][0]
                    p_02_idx = np.where(np.sum(points_refine == p_02, axis=1) == 3)[0][0]
                    tris_refine = np.vstack((tris_refine, np.array([[p_0_idx, p_01_idx, p_02_idx],
                                                                    [p_01_idx, p_2_idx, p_02_idx],
                                                                    [p_01_idx, p_1_idx, p_2_idx]])))

                elif p_on_01 and p_on_12 and not p_on_02:
                    ele_idx_del.append(ele_sur)
                    p_01_idx = np.where(np.sum(points_refine == p_01, axis=1) == 3)[0][0]
                    p_12_idx = np.where(np.sum(points_refine == p_12, axis=1) == 3)[0][0]
                    tris_refine = np.vstack((tris_refine, np.array([[p_0_idx, p_01_idx, p_2_idx],
                                                                    [p_01_idx, p_12_idx, p_2_idx],
                                                                    [p_01_idx, p_1_idx, p_12_idx]])))

        if verbose:
            print("Deleting old triangles ...")

        # delete old triangles
        ele_idx_roi = np.where(mask_roi)[0]
        ele_idx_lst_del = ele_idx_del + list(ele_idx_roi)
        tris_refine = np.delete(tris_refine, ele_idx_lst_del, 0)

        points_refine = np.round_(points_refine, 5)

        # # delete duplicate points
        # p_added = points_refine[points.shape[0]:, :]
        #
        # point_idx_del = np.array([])
        # for i_p, p in tqdm(enumerate(p_added)):
        #
        #     p_idx = np.where(np.sum(p == points_refine, axis=1) == 3)[0]
        #
        #     if len(p_idx) > 1:
        #         if p_idx[1] not in point_idx_del:
        #             point_idx_del = np.hstack((point_idx_del, p_idx[1:]))
        #
        #             # loop over point_idx_del and replace with first point idx
        #             for p_d_idx in p_idx[1:]:
        #                 tris_refine[tris_refine == p_d_idx] = p_idx[0]
        #
        # point_idx_keep = [i for i in range(points_refine.shape[0]) if i not in point_idx_del]
        # point_idx_new = [i for i in range(len(point_idx_keep))]
        # points_refine = points_refine[point_idx_keep, :]
        #
        # # renumber
        # for p_idx_keep, p_idx_new in zip(point_idx_keep[points.shape[0]:], point_idx_new[points.shape[0]:]):
        #     tris_refine[tris_refine == p_idx_keep] = p_idx_new

        # create new trimesh
        mesh = trimesh.Trimesh(vertices=points_refine,
                               faces=tris_refine)

        if repair:
            if mesh.is_watertight:
                if verbose:
                    print(f"Surface is watertight ...")
                mesh_ok = True
            else:
                if verbose:
                    print(f"Surface is NOT watertight ... trying to repair mesh ... ")
                # repair mesh
                trimesh.repair.fill_holes(mesh)

                if mesh.is_watertight:
                    if verbose:
                        print(f"Surface repaired ...")
                    mesh_ok = True

                else:
                    mesh_ok = False
                    radius -= 1
                    radius_ = radius + 2

                    if verbose:
                        print(f"WARNING: Could not repair refined surface ... "
                              f"shrinking radius by 1 mm to {radius} mm")
        else:
            mesh_ok = True

        if mesh_ok:
            if verbose:
                print(f"Saving {fn_surf_refined} ...")
            mesh.export(fn_surf_refined, file_type='stl_ascii')

            if remesh:
                # remesh surface
                print(f"Remeshing {fn_surf_refined} ...")
                command = f"meshfix {fn_surf_refined} -a 2.0 -u 1 -q --shells 9 " \
                          f"--stl -o {fn_surf_refined}"
                os.popen(command).read()

            refine = False


def map_data_to_surface(datasets, points_datasets, con_datasets, fname_fsl_gm, fname_fsl_wm, fname_midlayer=None,
                        delta=0.5, input_data_in_center=True, return_data_in_center=True, data_substitute=-1):
    """
    Maps data from ROI of fsl surface (wm, gm, or midlayer) to given Freesurfer brain surface (wm, gm, inflated).

    Parameters
    ----------
    datasets : np.ndarray of float [N_points x N_data] or list of np.ndarray
        Data in nodes or center of triangles in ROI (specify this in "data_in_center")
    points_datasets : np.ndarray of float [N_points x 3] or list of np.ndarray
        Point coordinates (x,y,z) of ROI where data in datasets list is given, the points have to be a subset of the
        GM/WM surface (has to be provided for each dataset)
    con_datasets : np.ndarray of int [N_tri x 3] or list of np.ndarray
        Connectivity matrix of dataset points (has to be provided for each dataset)
    fname_fsl_gm : str or list of str or list of None
        Filename of pial surface fsl file(s) (one or two hemispheres)
        e.g. in mri2msh: .../fs_ID/surf/lh.pial
    fname_fsl_wm : str or list of str or list of None
        Filename of wm surface fsl file(s) (one or two hemispheres)
        e.g. in mri2msh: .../fs_ID/surf/lh.white
    fname_midlayer : str or list of str
        Filename of midlayer surface fsl file(s) (one or two hemispheres)
        e.g. in headreco: .../fs_ID/surf/lh.central
    delta : float
        Distance parameter where gm-wm surface was generated 0...1 (default: 0.5)
        0 -> WM surface
        1 -> GM surface
    input_data_in_center : bool
        Flag if data in datasets in given in triangle centers or in points (Default: True)
    return_data_in_center : bool
        Flag if data should be returned in nodes or in elements (Default: True)
    data_substitute : float
        Data substitute with this number for all points in the inflated brain, which do not belong to the given data set

    Returns
    -------
    data_mapped : np.ndarray of float [N_points_inf x N_data]
        Mapped data to target brain surface. In points or elements
    """

    if type(fname_fsl_gm) is not list:
        fname_fsl_gm = [fname_fsl_gm]

    if type(fname_fsl_wm) is not list:
        fname_fsl_wm = [fname_fsl_wm]

    if type(fname_midlayer) is not list:
        fname_midlayer = [fname_midlayer]

    if fname_midlayer[0] is None:
        # load all freesurfer surfaces of gm and wm (hemispheres) and create midlayer
        points_gm = []
        con_target = []
        points_wm = []
        con_idx = 0

        for f_gm, f_wm in zip(fname_fsl_gm, fname_fsl_wm):
            p_gm, c_tar = nibabel.freesurfer.read_geometry(f_gm)
            p_wm, _ = nibabel.freesurfer.read_geometry(f_wm)

            points_gm.append(p_gm)
            points_wm.append(p_wm)
            con_target.append(c_tar + con_idx)
            con_idx += np.max(c_tar) + 1  # c_tar.shape[0]

        points_gm = np.vstack(points_gm)
        points_wm = np.vstack(points_wm)
        con_target = np.vstack(con_target)

        # regenerate the gm-wm surface w/o cropping in order to find congruent points
        wm_gm_vector = points_gm - points_wm

        # determine wm-gm surface (midlayer)
        points = points_wm + wm_gm_vector * delta

    else:
        # load directly all freesurfer midlayer surfaces (hemispheres)
        points = []
        con_target = []
        con_idx = 0

        for f_mid in fname_midlayer:
            if f_mid.endswith('.gii'):
                img = nibabel.gifti.giftiio.read(f_mid)
                p_mid = img.agg_data('pointset')
                c_tar = img.agg_data('triangle')
            else:
                p_mid, c_tar = nibabel.freesurfer.read_geometry(f_mid)

            points.append(p_mid)
            con_target.append(c_tar + con_idx)
            con_idx += np.max(c_tar) + 1  # c_tar.shape[0]

        points = np.vstack(points)
        con_target = np.vstack(con_target)

    # check datasets
    if type(datasets) is not list:
        datasets = [datasets]

    for i in range(len(datasets)):
        if datasets[i].ndim == 1:
            datasets[i] = datasets[i][:, np.newaxis]
        elif datasets[i].shape[0] < datasets[i].shape[1]:
            raise Warning("Datasets #{} shape[0] dimension is smaller than shape[1] (less points than dataset"
                          "components). Input dimension should be [N_points x N_data] ")

    if type(points_datasets) is not list:
        points_datasets = [points_datasets]

    if type(con_datasets) is not list:
        con_datasets = [con_datasets]
    # check if all points and all con are the same (if so, just map once and reuse results)
    all_points_equal = all([(points_datasets[i] == points_datasets[i + 1]).all()
                            for i in range(len(points_datasets) - 1)])

    all_con_equal = all([(con_datasets[i] == con_datasets[i + 1]).all()
                         for i in range(len(con_datasets) - 1)])

    if all_points_equal and all_con_equal:
        n_main_iter = 1
        n_sub_iter = len(datasets)
    else:
        n_main_iter = len(datasets)
        n_sub_iter = 1

    # check if indexation starts with value greater zero
    if np.min(con_target) > 0:
        con_target = con_target - np.min(con_target)

    n_points = points.shape[0]

    data_mapped = []

    # for i, data in enumerate(datasets):
    for i in range(n_main_iter):
        n_data = datasets[i].shape[1] if datasets[i].ndim > 1 else 1

        # n_points_cropped = points_datasets[i].shape[0]

        # check if indexation starts with value greater zero
        if np.min(con_datasets[i]) > 0:
            con_datasets[i] = con_datasets[i] - np.min(con_datasets[i])

        if datasets[i].ndim == 1:
            datasets[i] = datasets[i][:, np.newaxis]

        if input_data_in_center and return_data_in_center:
            # determine triangle center of dataset
            triangle_center_datasets = np.average(points_datasets[i][con_datasets[i]], axis=1)

            # determine triangle center of whole surface
            triangle_center_surface = np.average(points[con_target], axis=1)

            # loop over all points to get index list
            # point_idx_target = []
            # point_idx_data = []

            point_idx_target = np.zeros(datasets[i].shape[0])
            point_idx_data = np.arange(datasets[i].shape[0])
            for j in tqdm(range(datasets[i].shape[0]), desc="Mapping ROI to surface"):
                point_idx_target[j] = np.where(np.all(np.isclose(triangle_center_datasets[j,], triangle_center_surface),
                                                      axis=1))[0]
            point_idx_target = point_idx_target.astype(int).tolist()
            point_idx_data = point_idx_data.astype(int).tolist()

            # run subiterations (if all points and cons are equal, we save a lot of time here)
            for k in range(n_sub_iter):
                data_mapped.append(np.zeros([triangle_center_surface.shape[0], n_data]) + data_substitute * 1.0)
                data_mapped[k][point_idx_target, :] = datasets[k][point_idx_data, :]

        else:
            # loop over all points to get index list
            point_idx_target = []
            point_idx_data = list(range(datasets[i].shape[0]))

            for j in range(datasets[i].shape[0]):
                point_idx_target.append(np.where(np.all(np.isclose(points_datasets[i][j,], points), axis=1))[0])

            point_idx_target = [int(p) for p in point_idx_target]
            point_idx_data = [int(p) for p in point_idx_data]

            # run subiterations (if all points and cons are equal, we save a lot of time here)
            for k in range(n_sub_iter):
                # transform data from triangle center to triangle nodes if necessary
                if input_data_in_center:
                    data_nodes = data_elements2nodes(datasets[k], con_datasets[k])
                else:
                    data_nodes = datasets[k]

                # find and map data points
                data_mapped.append(np.zeros([n_points, n_data]) + data_substitute * 1.0)
                data_mapped[k][point_idx_target] = data_nodes[point_idx_data]

                # return data in elements instead of points
                if return_data_in_center:
                    data_mapped[k] = data_nodes2elements(data_mapped[k], con_target)

    return data_mapped


def midlayer_2_surf(midlayer_data, coords_target, coords_midlayer, midlayer_con=None, midlayer_data_in_nodes=False,
                    max_dist=5,
                    outside_roi_val=0, precise_map=True):
    """
    Convert midlayer data to whole-brain surface data, e.g. grey matter.
    Output is returned as data in nodes.

    Parameters
    ----------
    midlayer_data : np.ndarray of float
        (n_elm_midlayer,) or (n_nodes_midlayer,), the data in the midlayer.
    coords_target : np.ndarray of float
        (n_nodes_target, 3) Coordinates of the nodes of the target surface.
    coords_midlayer : np.ndarray of float
        (n_nodes_midlayer, 3) Coordinates of the nodes of the midlayer surface.
    midlayer_con : np.ndarray of int, optional
        (n_elm_midlayer, 3) Connectivity of the midlayer elements. Provide if data_in_points == True.
    midlayer_data_in_nodes : bool, default=False
        If midlayer data is provided in nodes, set to True and provide midlayer_con.
    max_dist : float, default=5
        Maximum distance between target and midlayer nodes to pull data from midlayer_data for.
    outside_roi_val : float, default=0
        Areas outside of max_dist are filled with outside_roi_val.
    precise_map : bool, default=True
        If elements to nodes mapping is done, perform this precise and slow or not.

    Returns
    -------
    data_target : np.ndarray
        (n_nodes_target, 1) The data in nodes of the target surface.
    """
    if not midlayer_data_in_nodes:
        assert midlayer_con is not None
        midlayer_data = np.squeeze(pynibs.data_elements2nodes(midlayer_data, midlayer_con, precise=precise_map))

    data_target = np.zeros((coords_target.shape[0]))
    for i in tqdm(range(data_target.shape[0]), desc='Mapping midlayer2surface'):
        idx = np.linalg.norm(coords_target[i] - coords_midlayer, axis=1).argmin()
        if np.linalg.norm(coords_target[i] - coords_midlayer[idx]) > max_dist:
            data_target[i] = outside_roi_val
        else:
            data_target[i] = midlayer_data[idx]

    return data_target


def point_data_to_cell_data_vtk(mesh=None, nodes=None, con=None, point_data=None, fn=None):
    """
    Convert point data to cell data in a VTK unstructured grid and save the result to a file.

    Parameters
    ----------
    mesh : meshio.Mesh, optional
        The mesh object containing points and cells.
    nodes : np.ndarray of float, optional
        (N_points, 3) Coordinates of the nodes.
    con : np.ndarray of int, optional
        (N_elements, 3) Connectivity index list forming the elements.
    point_data : dict, optional
        Point data to be transformed to cell data.
    fn : str, optional
        If provided, vtk file is written out to this file.

    Returns
    -------
    dict : cell_data
        All data sets from mesh transformed to cell_data.
    """
    if mesh is not None:
        assert nodes is None and con is None, "Provide either mesh or nodes and con."
    else:
        assert nodes is not None and con is not None, "Provide either mesh or nodes and con."
        mesh = meshio.Mesh(points=nodes, cells=[("triangle", con)], point_data=point_data)

    # Create VTK mesh instance from meshio object
    vtk_unstrgrid = vtkUnstructuredGrid()
    points = vtkPoints()
    number_of_points = mesh.points.shape[0]
    number_of_cells = mesh.cells[0].data.shape[0]
    cell_data = {}

    # Insert points into VTK points object
    for idx, p in tqdm(enumerate(mesh.points),
                       desc='Processing points', leave=False):
        points.InsertPoint(idx, p)
    vtk_unstrgrid.SetPoints(points)
    del points

    # Allocate and insert cells into VTK unstructured grid
    vtk_unstrgrid.Allocate(number_of_cells)
    assert len(mesh.cells) == 1, f"Only one cell block is supported, {len(mesh.cells)} present in mesh."
    for idx in tqdm(mesh.cells[0].data,
                    desc='Processing cells', leave=False):
        vtk_unstrgrid.InsertNextCell(VTK_TRIANGLE, 3, idx)

    for arr_name, arr_data_points_meshio in mesh.point_data.items():
        # Create and set point data array
        if len(arr_data_points_meshio.shape) == 1:
            arr_data_points_meshio = arr_data_points_meshio[:, np.newaxis]
        # arr_data_points_meshio = np.atleast_2d(arr_data_points_meshio)
        n_comps = arr_data_points_meshio.shape[1]
        arr_data_points_vtk = vtkDoubleArray()
        arr_data_points_vtk.SetNumberOfComponents(n_comps)
        arr_data_points_vtk.SetNumberOfTuples(number_of_points)
        arr_data_points_vtk.SetName(arr_name)

        for idx, data in tqdm(enumerate(arr_data_points_meshio), total=arr_data_points_meshio.shape[0],
                              desc=f"Processing data {arr_name}", leave=False):
            arr_data_points_vtk.SetTuple(idx, data)
        vtk_unstrgrid.GetPointData().AddArray(arr_data_points_vtk)

        # Convert point data to cell data
        p2c_conv = vtkPointDataToCellData()
        p2c_conv.SetInputData(vtk_unstrgrid)
        p2c_conv.Update()
        ptdata_unstrgrid = p2c_conv.GetOutput()
        arr_data_cell_vtk = ptdata_unstrgrid.GetCellData().GetArray(arr_name)

        # Add cell data array to the unstructured grid
        vtk_unstrgrid.GetCellData().AddArray(arr_data_cell_vtk)
        cell_data[arr_name] = numpy_support.vtk_to_numpy(arr_data_cell_vtk)

    # Write the VTK mesh to a file
    if fn is not None:
        write_vtu(fn, vtk_unstrgrid)

    return cell_data


def cell_data_to_point_data_vtk(mesh=None, nodes=None, con=None, cell_data=None, fn=None):
    """
    Convert cell data to point data in a VTK unstructured grid and save the result to a file.

    Parameters
    ----------
    mesh : meshio.Mesh
        The mesh object containing points and cells.
    nodes : np.ndarray of float, optional
        (N_points, 3) Coordinates of the nodes.
    con : np.ndarray of int, optional
        (N_elements, 3) Connectivity index list forming the elements.
    cell_data : dict, optional
        Cell data to be transformed to point data. keys: str, values: np.ndarray
    fn : str, optional
        If provided, vtk file is written out to this file.

    Returns
    -------
    dict : point_data
        All data sets from mesh transformed to point data.
    """
    if mesh is not None:
        assert nodes is None and con is None, "Provide either mesh or nodes and con."
    else:
        assert nodes is not None and con is not None, "Provide either mesh or nodes and con."
        mesh = meshio.Mesh(points=nodes, cells=[("triangle", con)], cell_data=cell_data)
    # Create VTK mesh instance from meshio object
    vtk_unstrgrid = vtkUnstructuredGrid()
    points = vtkPoints()
    # number_of_points = mesh.points.shape[0]
    number_of_cells = mesh.cells[0].data.shape[0]
    point_data = {}

    # Insert points into VTK points object
    for idx, p in tqdm(enumerate(mesh.points), desc='Processing points', leave=False):
        points.InsertPoint(idx, p)
    vtk_unstrgrid.SetPoints(points)
    del points

    # Allocate and insert cells into VTK unstructured grid
    vtk_unstrgrid.Allocate(number_of_cells)
    assert len(mesh.cells) == 1, f"Only one cell block is supported, {len(mesh.cells)} present in mesh."
    for idx in tqdm(mesh.cells[0].data, desc='Processing cells', leave=False):
        vtk_unstrgrid.InsertNextCell(VTK_TRIANGLE, 3, idx)

    for arr_name, arr_data_cells_meshio in mesh.cell_data.items():
        if isinstance(arr_data_cells_meshio, list):
            if len(arr_data_cells_meshio) > 1:
                raise ValueError
            arr_data_cells_meshio = arr_data_cells_meshio[0]
        # Create and set cell data array
        if len(arr_data_cells_meshio.shape) == 1:
            arr_data_cells_meshio = arr_data_cells_meshio[:, np.newaxis]
        # arr_data_cells_meshio = np.atleast_2d(arr_data_cells_meshio)
        n_comps = arr_data_cells_meshio.shape[1]
        arr_data_cells_vtk = vtkDoubleArray()
        arr_data_cells_vtk.SetNumberOfComponents(n_comps)
        arr_data_cells_vtk.SetNumberOfTuples(number_of_cells)
        arr_data_cells_vtk.SetName(f"{arr_name}")

        for idx, data in tqdm(enumerate(arr_data_cells_meshio), total=arr_data_cells_meshio.shape[1],
                              desc=f"Processing data {arr_name}", leave=False):
            arr_data_cells_vtk.SetTuple(idx, data)
        vtk_unstrgrid.GetCellData().AddArray(arr_data_cells_vtk)

        # Convert cell data to point data
        c2p_conv = vtkCellDataToPointData()
        c2p_conv.SetInputData(vtk_unstrgrid)
        c2p_conv.Update()
        ptdata_unstrgrid = c2p_conv.GetOutput()
        arr_data_point_vtk = ptdata_unstrgrid.GetPointData().GetArray(f"{arr_name}")

        # Add point data array to the unstructured grid
        vtk_unstrgrid.GetPointData().AddArray(arr_data_point_vtk)
        point_data[arr_name] = numpy_support.vtk_to_numpy(arr_data_point_vtk)

    # Write the VTK mesh to a file
    if fn is not None:
        write_vtu(fn, vtk_unstrgrid)

    return point_data


def write_vtu(fn, vtk_grid):
    writer = vtkXMLUnstructuredGridWriter()
    writer.SetFileName(fn)
    writer.SetInputData(vtk_grid)
    writer.Write()



import os
import numpy as np
from inspect import getsourcefile

import pynibs

# path of create_subject_XXXXX.py file (has to be in subjects root directoy
subject_folder = os.path.split(os.path.abspath(getsourcefile(lambda: 0)))[0]

# subject information
###########################################################################################################
subject_id = 'testsub'
fn_subject_obj = subject_folder + f'/{subject_id}.hdf5'

# mri information
###########################################################################################################
mri = [dict()]

mri[0]['fn_mri_T1'] = subject_folder + f'/mri/0/original/nifti/t1/2_T1_mprage_iso.nii.gz'
mri[0]['fn_mri_T2'] = subject_folder + f'/mri/0/original/nifti/t2/5_T2_cube_iso.nii.gz'
mri[0]['fn_mri_DTI'] = subject_folder + f'/mri/0/original/nifti/dti/7_HB4_diff_64dir_PA0_112.nii.gz'
mri[0]['fn_mri_DTI_rev'] = subject_folder + f'/mri/0/original/nifti/dti/8_HB4_diff_AP1_slice2.nii.gz'
mri[0]['fn_mri_DTI_bvec'] = subject_folder + f'/mri/0/original/nifti/dti/7_HB4_diff_64dir_PA0_112.bvec'
mri[0]['fn_mri_DTI_bval'] = subject_folder + f'/mri/0/original/nifti/dti/7_HB4_diff_64dir_PA0_112.bval'
mri[0]['dti_readout_time'] = '0.08214'
mri[0]['dti_phase_direction'] = 'y'

# mesh information
###########################################################################################################
mesh = {'charm_refm1_sm': dict(), 'headreco': dict(), 'headreco_refined_m1': dict()}

mesh['charm_refm1_sm']['info'] = 'charm --v 1.0 (simnibs 4.0 beta)'
mesh['charm_refm1_sm']['approach'] = 'charm'
mesh['charm_refm1_sm']['mesh_folder'] = os.path.join(subject_folder, 'testsub/mesh', 'charm_refm1_sm')
mesh['charm_refm1_sm']['vertex_density'] = None
mesh['charm_refm1_sm']['fn_mesh_msh'] = os.path.join(mesh['charm_refm1_sm']['mesh_folder'], f'm2m_{subject_id}',
                                                     f'{subject_id}.msh')
mesh['charm_refm1_sm']['fn_mesh_hdf5'] = os.path.join(mesh['charm_refm1_sm']['mesh_folder'], f'm2m_{subject_id}',
                                                      f'{subject_id}.hdf5')
mesh['charm_refm1_sm']['fn_tensor_vn'] = f'm2m_{subject_id}/DTI_coregT1_tensor.nii.gz'
mesh['charm_refm1_sm']['mri_idx'] = 0
mesh['charm_refm1_sm']['fn_mri_conform'] = f"m2m_{subject_id}/T1.nii.gz"

mesh['charm_refm1_sm']['fn_lh_wm'] = None
mesh['charm_refm1_sm']['fn_rh_wm'] = None
mesh['charm_refm1_sm']['fn_lh_gm'] = None
mesh['charm_refm1_sm']['fn_rh_gm'] = None
mesh['charm_refm1_sm']['fn_lh_gm_curv'] = None
mesh['charm_refm1_sm']['fn_rh_gm_curv'] = None
mesh['charm_refm1_sm']['fn_lh_midlayer'] = f'm2m_{subject_id}/surfaces/lh.central.gii'
mesh['charm_refm1_sm']['fn_rh_midlayer'] = f'm2m_{subject_id}/surfaces/rh.central.gii'
mesh['charm_refm1_sm']['smooth_skin'] = .8
mesh['charm_refm1_sm']['refinement_roi'] = 'midlayer_m1s1pmd'
mesh['charm_refm1_sm']['refinemement_element_size'] = 1

mesh['headreco']['info'] = 'headreco --v 1.0 (simnibs 3.1)'
mesh['headreco']['approach'] = 'headreco'
mesh['headreco']['mesh_folder'] = os.path.join(subject_folder, 'testsub/mesh', 'headreco')
mesh['headreco']['vertex_density'] = 1.0
mesh['headreco']['fn_mesh_msh'] = os.path.join(mesh['headreco']['mesh_folder'], f'{subject_id}.msh')
mesh['headreco']['fn_mesh_hdf5'] = os.path.join(mesh['headreco']['mesh_folder'], f'{subject_id}.hdf5')
mesh['headreco']['fn_tensor_vn'] = f'd2c_{subject_id}_PA/CTI_vn_tensor.nii'
mesh['headreco']['mri_idx'] = 0
mesh['headreco']['fn_lh_wm'] = None
mesh['headreco']['fn_rh_wm'] = None
mesh['headreco']['fn_lh_gm'] = None
mesh['headreco']['fn_rh_gm'] = None
mesh['headreco']['fn_lh_gm_curv'] = None
mesh['headreco']['fn_rh_gm_curv'] = None
mesh['headreco']['fn_lh_midlayer'] = f'fs_{subject_id}/surf/lh.central'
mesh['headreco']['fn_rh_midlayer'] = f'fs_{subject_id}/surf/rh.central'

mesh['headreco_refined_m1']['info'] = 'headreco --v 1.0 (simnibs 3.1)'
mesh['headreco_refined_m1']['approach'] = 'headreco'
mesh['headreco_refined_m1']['mesh_folder'] = os.path.join(subject_folder, 'testsub/mesh', 'headreco_refined_m1')
mesh['headreco_refined_m1']['vertex_density'] = 1.0
mesh['headreco_refined_m1']['center'] = [-41.33, -49.37, 73.78]
mesh['headreco_refined_m1']['radius'] = 35
mesh['headreco_refined_m1']['element_size'] = 0.4
mesh['headreco_refined_m1']['refine_domains'] = ["wm", "gm", "csf"]
mesh['headreco_refined_m1']['fn_mesh_msh'] = os.path.join(mesh['headreco_refined_m1']['mesh_folder'],
                                                          f"{subject_id}.msh")
mesh['headreco_refined_m1']['fn_mesh_hdf5'] = os.path.join(mesh['headreco_refined_m1']['mesh_folder'],
                                                           f"{subject_id}.hdf5")
mesh['headreco_refined_m1']['fn_tensor_vn'] = f"d2c_{subject_id}/dti_results_T1space/DTI_conf_tensor.nii.gz"
mesh['headreco_refined_m1']['mri_idx'] = 0
mesh['headreco_refined_m1']['fn_mri_conform'] = f"{subject_id}_T1fs_conform.nii.gz"
mesh['headreco_refined_m1']['smooth_domains'] = ['skin']

mesh['headreco_refined_m1']['fn_lh_wm'] = None
mesh['headreco_refined_m1']['fn_rh_wm'] = None
mesh['headreco_refined_m1']['fn_lh_gm'] = None
mesh['headreco_refined_m1']['fn_rh_gm'] = None
mesh['headreco_refined_m1']['fn_lh_gm_curv'] = None
mesh['headreco_refined_m1']['fn_rh_gm_curv'] = None
mesh['headreco_refined_m1']['fn_lh_midlayer'] = f"fs_{subject_id}/surf/lh.central"
mesh['headreco_refined_m1']['fn_rh_midlayer'] = f"fs_{subject_id}/surf/rh.central"

# roi information (first index: mesh, second index: roi)
###########################################################################################################
roi = {'headreco': {'midlayer_lh_rh': dict(), 'midlayer_m1s1pmd': dict()},
       'headreco_refined_m1': {'midlayer_lh_rh': dict(), 'midlayer_m1s1pmd': dict(),
                               'midlayer_m1s1pmd_refined': dict()},
       'charm_refm1_sm': {'midlayer_lh_rh': dict(), 'midlayer_m1s1pmd': dict()}}

roi['charm_refm1_sm']['midlayer_lh_rh']['type'] = 'surface'
roi['charm_refm1_sm']['midlayer_lh_rh']['info'] = 'freesurfer lh + rh whole brain midlayer'
roi['charm_refm1_sm']['midlayer_lh_rh']['gm_surf_fname'] = [mesh['charm_refm1_sm']['fn_lh_gm'],
                                                            mesh['charm_refm1_sm']['fn_rh_gm']]
roi['charm_refm1_sm']['midlayer_lh_rh']['wm_surf_fname'] = [mesh['charm_refm1_sm']['fn_lh_wm'],
                                                            mesh['charm_refm1_sm']['fn_rh_wm']]
roi['charm_refm1_sm']['midlayer_lh_rh']['midlayer_surf_fname'] = [mesh['charm_refm1_sm']['fn_lh_midlayer'],
                                                                  mesh['charm_refm1_sm']['fn_rh_midlayer']]
roi['charm_refm1_sm']['midlayer_lh_rh']['delta'] = 0.5
roi['charm_refm1_sm']['midlayer_lh_rh']['refine'] = False
roi['charm_refm1_sm']['midlayer_lh_rh']['X_ROI'] = None
roi['charm_refm1_sm']['midlayer_lh_rh']['Y_ROI'] = None
roi['charm_refm1_sm']['midlayer_lh_rh']['Z_ROI'] = None
roi['charm_refm1_sm']['midlayer_lh_rh']['layer'] = 3
roi['charm_refm1_sm']['midlayer_lh_rh']['fn_mask'] = None
roi['charm_refm1_sm']['midlayer_lh_rh']['fn_mask_avg'] = None

roi['charm_refm1_sm']['midlayer_m1s1pmd']['type'] = 'surface'
roi['charm_refm1_sm']['midlayer_m1s1pmd']['info'] = 'freesurfer PMd, M1 and somatosensory cortex'
roi['charm_refm1_sm']['midlayer_m1s1pmd']['gm_surf_fname'] = mesh['charm_refm1_sm']['fn_lh_gm']
roi['charm_refm1_sm']['midlayer_m1s1pmd']['wm_surf_fname'] = mesh['charm_refm1_sm']['fn_lh_wm']
roi['charm_refm1_sm']['midlayer_m1s1pmd']['midlayer_surf_fname'] = mesh['charm_refm1_sm']['fn_lh_midlayer']
roi['charm_refm1_sm']['midlayer_m1s1pmd']['delta'] = 0.5
roi['charm_refm1_sm']['midlayer_m1s1pmd']['refine'] = False
roi['charm_refm1_sm']['midlayer_m1s1pmd']['X_ROI'] = None
roi['charm_refm1_sm']['midlayer_m1s1pmd']['Y_ROI'] = None
roi['charm_refm1_sm']['midlayer_m1s1pmd']['Z_ROI'] = None
roi['charm_refm1_sm']['midlayer_m1s1pmd']['layer'] = 3
roi['charm_refm1_sm']['midlayer_m1s1pmd']['fn_mask'] = f"roi/midlayer_m1s1pmd/mask_{subject_id}m1pmdss.mgh"
roi['charm_refm1_sm']['midlayer_m1s1pmd']['fn_mask_avg'] = '/lefthandknob_M1S1PMd.overlay'
roi['charm_refm1_sm']['midlayer_m1s1pmd']['hemisphere'] = "lh"

roi['headreco']['midlayer_lh_rh']['type'] = 'surface'
roi['headreco']['midlayer_lh_rh']['info'] = 'freesurfer lh + rh whole brain midlayer'
roi['headreco']['midlayer_lh_rh']['gm_surf_fname'] = [mesh['headreco']['fn_lh_gm'], mesh['headreco']['fn_rh_gm']]
roi['headreco']['midlayer_lh_rh']['wm_surf_fname'] = [mesh['headreco']['fn_lh_wm'], mesh['headreco']['fn_rh_wm']]
roi['headreco']['midlayer_lh_rh']['midlayer_surf_fname'] = [mesh['headreco']['fn_lh_midlayer'],
                                                            mesh['headreco']['fn_rh_midlayer']]
roi['headreco']['midlayer_lh_rh']['delta'] = 0.5
roi['headreco']['midlayer_lh_rh']['refine'] = False
roi['headreco']['midlayer_lh_rh']['X_ROI'] = None
roi['headreco']['midlayer_lh_rh']['Y_ROI'] = None
roi['headreco']['midlayer_lh_rh']['Z_ROI'] = None
roi['headreco']['midlayer_lh_rh']['layer'] = 3
roi['headreco']['midlayer_lh_rh']['fn_mask'] = None
roi['headreco']['midlayer_lh_rh']['fn_mask_avg'] = None

roi['headreco']['midlayer_m1s1pmd']['type'] = 'surface'
roi['headreco']['midlayer_m1s1pmd']['info'] = 'freesurfer PMd, M1 and somatosensory cortex'
roi['headreco']['midlayer_m1s1pmd']['gm_surf_fname'] = mesh['headreco']['fn_lh_gm']
roi['headreco']['midlayer_m1s1pmd']['wm_surf_fname'] = mesh['headreco']['fn_lh_wm']
roi['headreco']['midlayer_m1s1pmd']['midlayer_surf_fname'] = mesh['headreco']['fn_lh_midlayer']
roi['headreco']['midlayer_m1s1pmd']['delta'] = 0.5
roi['headreco']['midlayer_m1s1pmd']['refine'] = False
roi['headreco']['midlayer_m1s1pmd']['X_ROI'] = None
roi['headreco']['midlayer_m1s1pmd']['Y_ROI'] = None
roi['headreco']['midlayer_m1s1pmd']['Z_ROI'] = None
roi['headreco']['midlayer_m1s1pmd']['layer'] = 3
roi['headreco']['midlayer_m1s1pmd']['fn_mask'] = f"roi/midlayer_m1s1pmd/mask_{subject_id}m1pmdss.mgh"
roi['headreco']['midlayer_m1s1pmd']['fn_mask_avg'] = '/masks/lefthandknob_M1S1PMd.overlay'
roi['headreco']['midlayer_m1s1pmd']['hemisphere'] = "lh"

roi_id = 'midlayer_lh_rh'
roi['headreco_refined_m1'][roi_id]['type'] = 'surface'
roi['headreco_refined_m1'][roi_id]['info'] = 'freesurfer lh + rh whole brain midlayer'
roi['headreco_refined_m1'][roi_id]['gm_surf_fname'] = [mesh['headreco_refined_m1']['fn_lh_gm'],
                                                       mesh['headreco_refined_m1']['fn_rh_gm']]
roi['headreco_refined_m1'][roi_id]['wm_surf_fname'] = [mesh['headreco_refined_m1']['fn_lh_wm'],
                                                       mesh['headreco_refined_m1']['fn_rh_wm']]
roi['headreco_refined_m1'][roi_id]['midlayer_surf_fname'] = [mesh['headreco_refined_m1']['fn_lh_midlayer'],
                                                             mesh['headreco_refined_m1']['fn_rh_midlayer']]
roi['headreco_refined_m1'][roi_id]['refine'] = False
roi['headreco_refined_m1'][roi_id]['delta'] = 0.5
roi['headreco_refined_m1'][roi_id]['X_ROI'] = None
roi['headreco_refined_m1'][roi_id]['Y_ROI'] = None
roi['headreco_refined_m1'][roi_id]['Z_ROI'] = None
roi['headreco_refined_m1'][roi_id]['layer'] = 3
roi['headreco_refined_m1'][roi_id]['fn_mask'] = None
roi['headreco_refined_m1'][roi_id]['fn_mask_avg'] = None

roi_id = 'midlayer_m1s1pmd'
roi['headreco_refined_m1'][roi_id]['type'] = 'surface'
roi['headreco_refined_m1'][roi_id]['info'] = 'freesurfer PMd, M1 and somatosensory cortex'
roi['headreco_refined_m1'][roi_id]['gm_surf_fname'] = mesh['headreco_refined_m1']['fn_lh_gm']
roi['headreco_refined_m1'][roi_id]['wm_surf_fname'] = mesh['headreco_refined_m1']['fn_lh_wm']
roi['headreco_refined_m1'][roi_id]['midlayer_surf_fname'] = mesh['headreco_refined_m1']['fn_lh_midlayer']
roi['headreco_refined_m1'][roi_id]['refine'] = False
roi['headreco_refined_m1'][roi_id]['delta'] = 0.5
roi['headreco_refined_m1'][roi_id]['X_ROI'] = None
roi['headreco_refined_m1'][roi_id]['Y_ROI'] = None
roi['headreco_refined_m1'][roi_id]['Z_ROI'] = None
roi['headreco_refined_m1'][roi_id]['layer'] = 3
roi['headreco_refined_m1'][roi_id]['fn_mask'] = f"roi/midlayer_m1s1pmd/mask_{subject_id}m1pmdss.mgh"
roi['headreco_refined_m1'][roi_id]['fn_mask_avg'] = '//masks/lefthandknob_M1S1PMd.overlay'
roi['headreco_refined_m1'][roi_id]['hemisphere'] = "lh"

roi_id = 'midlayer_m1s1pmd_refined'
roi['headreco_refined_m1'][roi_id]['type'] = 'surface'
roi['headreco_refined_m1'][roi_id]['info'] = 'freesurfer PMd, M1 and somatosensory cortex'
roi['headreco_refined_m1'][roi_id]['gm_surf_fname'] = mesh['headreco_refined_m1']['fn_lh_gm']
roi['headreco_refined_m1'][roi_id]['wm_surf_fname'] = mesh['headreco_refined_m1']['fn_lh_wm']
roi['headreco_refined_m1'][roi_id]['midlayer_surf_fname'] = mesh['headreco_refined_m1']['fn_lh_midlayer']
roi['headreco_refined_m1'][roi_id]['refine'] = True
roi['headreco_refined_m1'][roi_id]['delta'] = 0.5
roi['headreco_refined_m1'][roi_id]['X_ROI'] = None
roi['headreco_refined_m1'][roi_id]['Y_ROI'] = None
roi['headreco_refined_m1'][roi_id]['Z_ROI'] = None
roi['headreco_refined_m1'][roi_id]['layer'] = 3
roi['headreco_refined_m1'][roi_id]['fn_mask'] = f"roi/midlayer_m1s1pmd/mask_{subject_id}m1pmdss.mgh"
roi['headreco_refined_m1'][roi_id]['fn_mask_avg'] = '//masks/lefthandknob_M1S1PMd.overlay'
roi['headreco_refined_m1'][roi_id]['hemisphere'] = "lh"

# experiment information
###########################################################################################################
exp = {'TMS_localite': dict(), 'TMS_ant': dict(), 'TMS_brainsight': dict()}

exp['TMS_localite']['info'] = ['TMS-MEP template Localite (MPI)']
exp['TMS_localite']['date'] = ['XX/XX/20XX']
exp['TMS_localite']['nnav_system'] = 'Localite'
exp['TMS_localite']['fn_tms_nav'] = [[subject_folder + '...xml', subject_folder + '...xml']]
exp['TMS_localite']['fn_data'] = [[subject_folder + '/exp/0/mep/...cfs']]
exp['TMS_localite']['fn_intensity'] = None
exp['TMS_localite']['mep_onsets'] = [[0]]
exp['TMS_localite']['fn_exp_csv'] = [subject_folder + '/exp/0/experiment.csv']
exp['TMS_localite']['fn_exp_hdf5'] = [subject_folder + '/exp/0/experiment.hdf5']
exp['TMS_localite']['fn_coil'] = [['.../MagStim_Alpha_Coil_70mm_Fig8.nii.gz']]
exp['TMS_localite']['fn_mri_nii'] = [[subject_folder + '/exp/0/..._conform.nii']]
exp['TMS_localite']['cond'] = [['']]
exp['TMS_localite']['tms_pulse_time'] = None
exp['TMS_localite']['experimenter'] = ''
exp['TMS_localite']['mep_fit_info'] = ['']
exp['TMS_localite']['incidents'] = ['']
exp['TMS_localite']['postproc'] = [dict()]
exp['TMS_localite']['postproc'][0]['info'] = ' '
exp['TMS_localite']['postproc'][0]['cmd'] = ' '

exp['TMS_ant']['info'] = ['TMS-MEP-EEG template (ANT)']
exp['TMS_ant']['date'] = ['XX/XX/20XX']
exp['TMS_ant']['nnav_system'] = 'Visor'
exp['TMS_ant']['fn_data'] = [[subject_folder + '/exp/2/...cnt']]
exp['TMS_ant']['fn_exp_hdf5'] = [subject_folder + '/exp/2/experiment.hdf5']
exp['TMS_ant']['fn_coil'] = [['.../MagStim_Alpha_Coil_70mm_Fig8.nii.gz']]
exp['TMS_ant']['fn_mri_nii'] = [[subject_folder + '/exp/2/..._conform.nii']]
exp['TMS_ant']['fn_visor_cnt'] = [subject_folder + '/exp/2/...pos.cnt']
exp['TMS_ant']['fn_visor_calibration'] = [subject_folder + '/exp/2/...calibration.xml']
exp['TMS_ant']['fn_eeg_cnt'] = [subject_folder + '/exp/2/eeg/...cnt']
exp['TMS_ant']['eeg_channels'] = ['all']
exp['TMS_ant']['fn_emg_cnt'] = [subject_folder + '/exp/2/emg/...cnt']
exp['TMS_ant']['emg_channels'] = [0]
exp['TMS_ant']['emg_max_duration'] = [10]
exp['TMS_ant']['eeg_max_duration'] = [10]
exp['TMS_ant']['emg_trigger_value'] = ['4']
exp['TMS_ant']['eeg_trigger_value'] = ['1']
exp['TMS_ant']['cond'] = [['M1_random']]
exp['TMS_ant']['tms_pulse_time'] = None
exp['TMS_ant']['experimenter'] = ' '
exp['TMS_ant']['mep_fit_info'] = [' ']
exp['TMS_ant']['incidents'] = [' ']
exp['TMS_ant']['fiducal_corr'] = [[0, 0, 0]]  # x,y,z in mm

exp['TMS_brainsight']['info'] = ['TMS-MEP-EEG template (Brainsight)']
exp['TMS_brainsight']['date'] = ['09/03/20222']
exp['TMS_brainsight']['nnav_system'] = 'Brainsight'
exp['TMS_brainsight']['fn_data'] = [subject_folder + '/Data_testsub.txt']
exp['TMS_brainsight']['fn_exp_hdf5'] = [subject_folder + '/exp/random/experiment.hdf5']
exp['TMS_brainsight']['fn_coil'] = [['.../Magstim_D70_AFC_force_sensor.ccd.nii.gz']]
exp['TMS_brainsight']['fn_mri_nii'] = [[subject_folder + '/exp/random/testsub_T1.nii.gz']]
exp['TMS_brainsight']['cond'] = [['M1_random']]
exp['TMS_brainsight']['tms_pulse_time'] = None
exp['TMS_brainsight']['experimenter'] = 'Zijian'
exp['TMS_brainsight']['mep_fit_info'] = [' ']
exp['TMS_brainsight']['incidents'] = [' ']
exp['TMS_brainsight']['fiducal_corr'] = [[0, 0, 0]]  # x,y,z in mm

# add plot settings
###########################################################################################################
ps = []

ps.append(pynibs.create_plot_settings_dict(plotfunction_type='surface_vector_plot'))
ps[0]['info'] = ' '
ps[0]['plot_function'] = 'surface_vector_plot'
ps[0]['fname_in'] = '..._data.xdmf'
ps[0]['fname_png'] = '...png'
ps[0]['png_resolution'] = 1.0
ps[0]['quantity'] = [' ']
ps[0]['vlabels'] = None
ps[0]['vscales'] = None
ps[0]['datarange'] = [None, None]
ps[0]['domain_label'] = 'tissue_type'
ps[0]['domain_IDs'] = 0
ps[0]['colorbar_label'] = ' '
ps[0]['colorbar_position'] = [0.45, -0.05]
ps[0]['colorbar_orientation'] = 'Horizontal'
ps[0]['colorbar_aspectratio'] = 20
ps[0]['colorbar_titlefontsize'] = 12
ps[0]['colorbar_labelfontsize'] = 12
ps[0]['colorbar_labelformat'] = '%-#6.1f'
ps[0]['colorbar_numberoflabels'] = 5
ps[0]['colorbar_labelcolor'] = [0, 0, 0]
ps[0]['view'] = [[-332.040630013204, -16.8948837495393, 256.721941018442],
                 [-35.3111635316966, -13.6999528376981, 36.4196465112876],
                 [0.588834647614276, -0.167617710066906, 0.790682022712606],
                 [30, 0, 0]]
ps[0]['interpolate'] = True
ps[0]['edges'] = False
ps[0]['axes'] = False
ps[0]['colormap'] = 'Rainbow Blended White'
ps[0]['opacitymap'] = None
ps[0]['vscale_mode'] = ['off']
ps[0]['vector_mode'] = {'Every Nth Point': 10}
ps[0]['vcolor'] = None
ps[0]['background_color'] = np.array([1, 1, 1])
ps[0]['viewsize'] = np.array([1000, 1000])

# save subject information in .hdf5 file
###########################################################################################################
pynibs.save_subject_hdf5(fname=fn_subject_obj,
                         mri_dict=mri,
                         mesh_dict=mesh,
                         roi_dict=roi,
                         exp_dict=exp,
                         ps_dict=ps,
                         subject_id=subject_id,
                         subject_folder=subject_folder
                         )

print(f'Created subject .hdf5 file: {fn_subject_obj}')



import unittest
import pynibs


class TestRegIonOfInterestSurface(unittest.TestCase):

    def test_regionofinterestsurface(self):
        roi = pynibs.RegionOfInterestSurface()
        for field in ["X_ROI", "Y_ROI", "Z_ROI"]:
            assert hasattr(roi, field)
        for field in ["x_roi", "x_roi", "x_roi"]:
            assert not hasattr(roi, field)


class TestRegionOfInterestVolume(unittest.TestCase):
    def test_regionofinterestvolume(self):
        roi = pynibs.RegionOfInterestVolume()
